"""Defines generic class for dataflow analysis to find constraints on transaction fields.

Possible values for a field are considered to be a set, referred to as universal set `U` for that field.
if U is finite and small, values are enumerated and are stored in the context. However, in case U is large,
such as address type fields, enum type values are used to represent UniversalSet and NullSet.

For a given `key` and a `basic_block`, block_contexts[key][basic_block] are values(V), such that, if the transaction field represented by the `key`
is set to one of the values present in V, then:
    - The execution might reach the `basic_block`
    - The execution might successfully reach the end of the `basic_block`
    - The execution might reach a leaf basic block which results in successful completion of execution.

block_contexts is computed in three steps, each step making the information more precise.
1: In the first step, local information is considered. For each block, information inferred from the instructions present in the
    block is computed.
        - if the basic block contains instructions `assert(txn OnCompletion == int UpdateApplication)`, then block context of this block
        for transaction types will be equal to `{ApplUpdateApplication}`.
        - if basic block errors, contains `err` instruction or `return 0`, then block context will be NullSet.
        - if instructions in the block does not provide any information related to the field, then block context will be equal to
        UniversalSet (all possible values) for that key.
2: In the second step, information from the predecessors is considered. For this, forward dataflow analysis is used.
    This problem is analogous to reaching definitions problem:
        I   Each possible value is a definition that is defined at the start of execution i.e defined at the start of entry block.
        II  The definition(value) will reach start of the basic block, if it reaches the end of one of it's predecessors.
        III The definition(value) will reach the end of the block, if it is preserved by the basic block or it is defined in the basic block.
        IV  No definition(value) is defined in a basic block
        V   The definition(value) will reach start of the basic block, if the condition specific to reach this block is satisfied. Condition used
            to determine the branch taken can contain constraints related to the analysis. Developers can branch to error block or success block
            based on the transaction field value. Path based context is used to combine this information in forward analysis.

    Equations:
        initialization:
            RCHin(entry) = UniversalSet()  - from (I)
            RCHout(b) = NullSet() for all basic blocks b.
        fixed point iteration:
            RCHin(b) = union(intersection(RCHout(prev_b), path_context[b][prev_b]) for prev_b in predecessors) - from (II), (V)
            RCHout(b) = union(GEN(b), intersection(RCHin(b), PRSV(b)) - from (III)
            GEN(b) = NullSet() for all basic blocks b.
            RCHout(b) = intersection(RCHin(b), PRSV(b))

    `PRSV(b)` is `block_contexts` computed in the first step.
    Reverse postorder ordering is used for the iteration algorithm.

3: Finally, information from the successors is combined using backward dataflow analysis similar to Live-variable analysis.
    - `block_contexts` is equal to reach out information computed in the second step.
    - For leaf blocks, value is live if the value is preserved by the block.
    - For other blocks, the value is live, if the value is used(preserved) by one of the successors.
    equations:
        initialization:
            LIVEout(b) = NullSet() for all non-leaf blocks.
            LIVEout(b) = PRSV(b) for all leaf blocks.
        fixed point iteration:
            LIVEin(b) = union(LIVEou(succ_b) for succ_b in successors)
            LIVEout(b) = intersection(LIVEin(b), PRSV(b))

    `PRSV(b)` is `block_contexts` after the second step.
    Postorder ordering is used for the iteration algorithm.

Blocks containing `callsub` instruction and blocks which are right after the `callsub` instruction are
treated differently.
e.g
    main:           // Basic Block B1
        int 2
        retsub

    path_1:         // Basic Block B2
        txn OnCompletion
        int UpdateApplication
        ==
        assert
        callsub main

        byte "PATH_1"      // Basic Block B3
        int 1
        return

    path_2:                 // Basic Block B4
        txn OnCompletion
        int DeleteApplication
        ==
        assert
        callsub main

        byte "PATH_2"        // Basic Block B5
        int 1
        return

CFG:
                                  B2
                                  +--------------------------+
                                  |        4: path_1:        |
                                  |   5: txn OnCompletion    |
                                  | 6: int UpdateApplication |
                                  |          7: ==           |
                                  |        8: assert         |
                                  |     9: callsub main      |
                                  +--------------------------+
                                    |
                                    |
B4                                  v   B1                          B5
+---------------------------+     +--------------------------+     +-------------------+
|        13: path_2:        |     |                          |     |                   |
|   14: txn OnCompletion    |     |         1: main:         |     | 19: byte "PATH_2" |
| 15: int DeleteApplication |     |         2: int 2         |     |     20: int 1     |
|          16: ==           |     |        3: retsub         |     |    21: return     |
|        17: assert         |     |                          |     |                   |
|     18: callsub main      | --> |                          | --> |                   |
+---------------------------+     +--------------------------+     +-------------------+
                                    |
                                    |
                                    v   B3
                                  +--------------------------+
                                  |    10: byte "PATH_1"     |
                                  |        11: int 1         |
                                  |        12: return        |
                                  +--------------------------+

CFG will have edges:
    - B2 -> B1       # callsub instruction transfers the execution to called subroutine
    - B4 -> B1
    - B1 -> B3       # execution returns to the next instruction present after callsub instruction
    - B1 -> B5

B3 and B5 are return points of the subroutine.
B3 is only executed if execution reaches B2 => block_context for txn type is `{DeleteApplication}`.
similarly, B5 is only executed if execution reaches B4 => block_context for txn type is `{UpdateApplication}`.

block_contexts["TransactionType"][B1] = `{UpdateApplication, DeleteApplication}` # from B2 -> B1 and B4 -> B1.

B3 and B5 are return points and have only one predecessor(B1) in CFG. As a result, block_contexts will be
block_contexts["TransactionType"][B4] = `{UpdateApplication, DeleteApplication}`
block_contexts["TransactionType"][B5] = `{UpdateApplication, DeleteApplication}`.

This is because, when traversing the CFG without differentiating subroutine blocks and others, possible execution paths will be:
1. B2 -> B1 -> B3
2. B4 -> B1 -> B5
3. B2 -> B1 -> B5  # won't be possible at runtime.
4. B4 -> B1 -> B3  # won't be possible at runtime.

However, At runtime, execution will reach B3 if and only if it reaches B2, same for B5 and B3. Using this reasoning while
combining information from predecessors and successors will give more accurate results.

---------------------------------------------------------------------------

`DataflowTransactionContext._get_asserted_single(key, ins_stack_value)` returns a Tuple of true_values and false_values.
    true_values: Given that the result of `ins_stack_value` is NON-ZERO, what are the possible
                values for the key.
    false_values: Given that the result of `ins_stack_value` is ZERO, what are the possible values
                for the key.

`_get_asserted_single` uses pattern matching to identify comparisons on interested fields. It only works when the
`ins_stack_value` directly represents the value of the comparison.
e.g
    - Eq(txn RekeyTo, global Zeroaddress)
    - Neq(txn OnCompletion, int UpdateApplication)
    - ...

But when the result depends on multiple equations it is NOT possible to use the same kind of pattern matching and derive
possible values.
e.g
ins_stack_value =   Or(
                        And(
                            Eq(txn RekeyTo, global ZeroAddress),
                            Eq(txn CloseRemainderTo, global ZeroAddress),
                        ),
                        And(
                            Eq(txn Fee, global MinTxnFee),
                            Eq(txn AssetCloseTo, global ZeroAddress),
                        ),
                    )


We can combine information from multiple equations when they are connected using And, Or, !.
e.g
    txn RekeyTo
    global ZeroAddress
    ==                      // equation <1>
    txn CloseRemainderTo
    global ZeroAddress
    ==                      // <2>
    &&
    txn Fee
    global MinTxnFee
    <=                      // <3>
    &&

=> And(
    And(
        Eq(txn RekeyTo, global ZeroAddress),
        Eq(txn CloseRemainderTo, global ZeroAddress),
    ),
    LessE(txn Fee, global MinTxnFee),
)

=> And(And(<1>, <2>), <3>)
=> ins_stack_value = And(<1>, <2>, <3>)           // flattend

if it is given that ins_stack_value is False, Then atleast ONE of the equations <1>, <2> or <3> is False.
if it is given that ins_stack_value is True, Then ALL of the equations <1>, <2> and <3> are True.

This applies for any number of equations: And(<1>, <2>, <3>, <4>, ....)

-> if the ins_stack_value is of form And(<1>, <2>, <3>, <4>, ....) and is `True`, Then we can combine possible values
from multiple equations using set `Intersection`.
-> if the ins_stack_value is of form And(<1>, <2>, <3>, <4>, ....) and is `False`, Then we can combine possible values
from multiple equations using set `Union`.

if any of equation in And(<1>, ...) is UnknownStackValue then `false_values` for And(...) is universal_set U.
    -> The result of And() could be false because the UnknownStackValue is False. None of the known values have
        to be False.
    -> true_values are not affected by having an UnknownStackValue. All the known values have to be True irrespective
        of unknown values for the result to be True.

Similar reasoning can be done for Or(<1>, <2>, <3>, <4>, ....)

if ins_stack_value is True, Then atleast ONE of the equations <1>, <2>, <3>, ... is True.   (Union)
if ins_stack_value is False, Then ALL of the equations <1>, <2>, <3>, ... are False.        (Intersection)

if any of equation in Or(<1>, ...) is UnknownStackValue then `true_values` for Or(...) is universal_set U.
    -> The result of Or() could be True because the UnknownStackValue is True. None of the known values have
        to be True.
    -> false_values are not affected by having an UnknownStackValue. All the known values have to be False irrespective
        of unknown values for the result to be False
"""


from abc import ABC, abstractmethod
import logging
from typing import TYPE_CHECKING, Any, Dict, List, Tuple, Set
from collections import defaultdict

from tealer.analyses.dataflow.transaction_context.utils.key_helpers import (
    get_gtxn_at_index_key,
    get_absolute_index_key,
    get_relative_index_key,
)
from tealer.teal.instructions.instructions import (
    Assert,
    Return,
    BZ,
    BNZ,
    Err,
    And,
    Or,
    Not,
    TealerCustomErrInstruction,
)
from tealer.utils.analyses import (
    is_int_push_ins,
    prev_blocks_global,
    next_blocks_global,
    leaf_block_global,
)
from tealer.utils.algorand_constants import MAX_GROUP_SIZE
from tealer.analyses.utils.stack_ast_builder import (
    KnownStackValue,
    UnknownStackValue,
    get_stack_value_for_ins,
    compute_equations,
)

if TYPE_CHECKING:
    from tealer.teal.teal import Teal
    from tealer.teal.functions import Function
    from tealer.teal.basic_blocks import BasicBlock
    from tealer.teal.instructions.instructions import Instruction


class IncorrectDataflowTransactionContextInitialization(Exception):
    pass


logger_txn_ctx = logging.getLogger("TransactionCtxAnalysis")
logging.basicConfig(level=logging.DEBUG)

debug_keys = ["TransactionType"]


class DataflowTransactionContext(ABC):  # pylint: disable=too-few-public-methods

    # List of keys, unique and separate context is stored for each key.
    # each key represents a transaction field.
    BASE_KEYS: List[str] = []
    # BASE_KEYS for which transaction context information from `gtxn {i} {field}` is also stored.
    KEYS_WITH_GTXN: List[str] = []  # every key in this list should also present in BASE_KEYS.

    def __init__(self, function: "Function"):
        # self._teal: "Teal" = teal
        self._function: "Function" = function
        # entry block of CFG
        self._entry_block: "BasicBlock" = function.entry
        # self._block_contexts[KEY][B] -> block_context of KEY for block B
        self._block_contexts: Dict[str, Dict["BasicBlock", Any]] = defaultdict(dict)
        # self._path_contexts[KEY][Bi][Bj] -> path_context of KEY for path Bj -> Bi
        self._path_contexts: Dict[str, Dict["BasicBlock", Dict["BasicBlock", Any]]] = defaultdict(
            dict
        )
        if not self.BASE_KEYS:
            raise IncorrectDataflowTransactionContextInitialization(
                f"BASE_KEYS are not initialized {self.__class__.__name__}"
            )

    @abstractmethod
    def _universal_set(self, key: str) -> Any:
        """Return universal set for the field corresponding to given key

        Args:
            key: The key used to represent values of a field in the analysis.

        Returns:
            Returns the universal set: list of all possible values for the field tracked by the
            :key:
        """

    @abstractmethod
    def _null_set(self, key: str) -> Any:
        """Return null set for the field corresponding to given key

        Args:
            key: The key used to represent values of a field in the analysis.

        Returns:
            Returns the null set: empty value for the field tracked by the :key:
        """

    @abstractmethod
    def _union(self, key: str, a: Any, b: Any) -> Any:
        """return union of a and b, where a, b represent values for the given key

        Args:
            key: The analysis key. The values in set :a: and :b: are values of this key.
            a: Set 1.
            b: Set 2.

        Returns:
            Returns union of set a and b.
        """

    @abstractmethod
    def _intersection(self, key: str, a: Any, b: Any) -> Any:
        """return intersection of a and b, where a, b represent values for the given key

        Args:
            key: The analysis key. The values in set :a: and :b: are values of this key.
            a: Set 1.
            b: Set 2.

        Returns:
            Returns union of set a and b.
        """

    @abstractmethod
    def _get_asserted_single(self, key: str, ins_stack_value: KnownStackValue) -> Tuple[Any, Any]:
        """For the given key and ins_stack, return true_values and false_values

        true_values: Given that the result of `ins_stack_value` is NON-ZERO, what are the possible
            values for the key.
        false_values: Given that the result of `ins_stack_value` is ZERO, what are the possible values
            for the key.

        Args:
            key: The key used to represent values of a field in the analysis.
            ins_stack_value: The stack value being considered for assertion. The :ins_stack_value: is
                not the result of "And" or "Or" or "Not" instructions.

        Returns:
            Returns the "true_values" and "false_values" for the key.
        """

    @abstractmethod
    def _store_results(self) -> None:
        """Store the collected information in the context object of each block"""

    def _get_asserted(self, key: str, ins_stack_value: KnownStackValue) -> Tuple[Any, Any]:
        """For the given key and ins_stack_value, return true_values and false_values

        true_values: Given that the result of `ins_stack_value` is NON-ZERO, what are the possible
            values for the key.
        false_values: Given that the result of `ins_stack_value` is ZERO, what are the possible values
            for the key.

        Args:
            key: The key used to represent values of a field in the analysis.
            ins_stack_value: The stack value being considered for assertion.

        Returns:
            Returns the "true_values" and "false_values" for the key.
        """
        if not isinstance(ins_stack_value.instruction, (And, Or, Not)):
            # and not isinstance(ins_stack_value.instruction, Or):
            # single equation
            t, f = self._get_asserted_single(key, ins_stack_value)
            return t, f
        if isinstance(ins_stack_value.instruction, Not):
            arg = ins_stack_value.args[0]
            if isinstance(arg, UnknownStackValue):
                # unknown value
                return self._universal_set(key), self._universal_set(key)
            true_values, false_values = self._get_asserted(key, arg)
            # swap the values
            final_true_values, final_false_values = false_values, true_values
            return final_true_values, final_false_values
        if isinstance(ins_stack_value.instruction, And):
            # x = And(<1>, <2>, <3>, ...),
            #   if x is True then <1>, <2>, ... are all True, Intersection
            #   if x is False then one of <1>, <2>, ... is False, Union
            individual_equations, has_unknown_value = compute_equations(ins_stack_value, And)
            final_true_values = self._universal_set(key)
            final_false_values = self._null_set(key)
            for equation in individual_equations:
                # combine values recursively.
                # And(<Eq()>, <Or()>, <Or()>, ...):
                #      values Or() cannot be calculated using _get_asserted_single
                true_values, false_values = self._get_asserted(key, equation)
                final_true_values = self._intersection(key, final_true_values, true_values)
                final_false_values = self._union(key, final_false_values, false_values)
            if has_unknown_value:
                # has_unknown_value is True if result of And depends on an unknown value.
                # if And is False, the unknown value could be false and that unknown value
                # might or might not depend on the key.
                # note: Having a unknown value does not affect true_values, known values have
                # to be True irrespective of unknown value for the result to be True
                final_false_values = self._universal_set(key)
            return final_true_values, final_false_values
        # x = Or(<1>, <2>, <3>, ....),
        #   if x is False then <1>, <2>, ... are all False,     Intersection
        #   if x is True then one of <1>, <2>, ... is True,     Union
        individual_equations, has_unknown_value = compute_equations(ins_stack_value, Or)
        final_false_values = self._universal_set(key)
        final_true_values = self._null_set(key)
        for equation in individual_equations:
            true_values, false_values = self._get_asserted(key, equation)
            final_false_values = self._intersection(key, final_false_values, false_values)
            final_true_values = self._union(key, final_true_values, true_values)
        if has_unknown_value:
            # has_unknown_value is True if result of Or depends on an unknown value.
            # if Or is True, the unknown value could be True and that unknown value
            # might or might not depend on the key
            # note: Having a unknown value does not affect false_values, known values have
            # to be False irrespective of unknown value for the result to be False
            final_true_values = self._universal_set(key)
        return final_true_values, final_false_values

    def _block_level_constraints(  # pylint: disable=too-many-branches
        self, analysis_keys: List[str], block: "BasicBlock"
    ) -> None:
        """Calculate and store constraints on keys applied within the block.

        By default, no constraints are considered i.e values are assumed to be universal_set
        if block contains `Err` or `Return 0`, values are set to null set.

        if block contains assert instruction, values are further constrained using the comparison being
        asserted. Values are stored in self._block_contexts
        self._block_contexts[KEY][B] -> block_context of KEY for block B

        Args:
            analysis_keys: List of keys, each representing a field.
            block: The basic block.

        Fills the _block_contexts[:key:][:block:] based on assertions performed by the instructions in the
        :block:
        """
        for key in analysis_keys:
            self._block_contexts[key][block] = self._universal_set(key)

        for ins in block.instructions:
            if isinstance(ins, Assert):
                assert_ins_stack_value = get_stack_value_for_ins(ins)
                assert_ins_arg = assert_ins_stack_value.args[0]
                if isinstance(assert_ins_arg, UnknownStackValue):
                    continue
                for key in analysis_keys:
                    # asserted_values, _ = self._get_asserted(key, stack)
                    asserted_values, _ = self._get_asserted(key, assert_ins_arg)
                    present_values = self._block_contexts[key][block]
                    self._block_contexts[key][block] = self._intersection(
                        key, present_values, asserted_values
                    )
            elif isinstance(ins, Return):
                # if return 0, set possible values to NullSet()
                return_ins_value = get_stack_value_for_ins(ins)
                return_ins_arg = return_ins_value.args[0]
                if isinstance(return_ins_arg, UnknownStackValue):
                    continue
                # if int 0; return; set possible values to null set.
                is_int, value = is_int_push_ins(return_ins_arg.instruction)
                if is_int and value == 0:
                    for key in analysis_keys:
                        self._block_contexts[key][block] = self._null_set(key)
                    continue
                # if And(<1>, <2>, ..); return; i.e return value depends on the result of a comparison
                for key in analysis_keys:
                    true_values, _ = self._get_asserted(key, return_ins_arg)
                    present_values = self._block_contexts[key][block]
                    self._block_contexts[key][block] = self._intersection(
                        key, present_values, true_values
                    )
            elif isinstance(ins, (Err, TealerCustomErrInstruction)):
                # if err, set possible values to NullSet()
                for key in analysis_keys:
                    self._block_contexts[key][block] = self._null_set(key)

    def _path_level_constraints(  # pylint: disable=too-many-branches
        self, analysis_keys: List[str], block: "BasicBlock"
    ) -> None:
        """Calculate and store constraints on keys applied along each path.

        By default, no constraints are considered i.e values are assumed to be universal_set

        if block contains bz/bnz instruction, possible values are calculated for each branch and
        are stored in self._path_contexts
        self._path_contexts[KEY][Bi][Bj] -> path_context of KEY for path Bj -> Bi

        Args:
            analysis_keys: List of keys, each representing a field.
            block: The basic block.

        Fills the _path_contexts[key][Bj][Bi] where :key: is the key in analysis_keys and
        Bi is the :block: and Bj are next blocks of :block: in the global CFG.
        """

        for key in analysis_keys:
            path_context = self._path_contexts[key]
            for b in next_blocks_global(self._function, block):
                # print(
                #     f"block: {repr(block)}, b: {repr(b)}, {key:}, block: {block.__hash__()}, b: {b.__hash__()}, block sub: {block.subroutine.name}, b sub: {b.subroutine.name}"
                # )
                # path_context[bi][bj]: path context of path bj -> bi, bi is the successor
                if b not in path_context:
                    path_context[b] = {}
                    # print(f"block: {repr(block)}, b: {repr(b)}, {key:}, ddd")

                # if there are no constraints, set the possible values to universal set
                path_context[b][block] = self._universal_set(key)

        if isinstance(block.exit_instr, (BZ, BNZ)):
            exit_ins_stack_value = get_stack_value_for_ins(block.exit_instr)
            exit_ins_arg = exit_ins_stack_value.args[0]
            if isinstance(exit_ins_arg, UnknownStackValue):
                return
            for key in analysis_keys:
                # true_values: possible values for {key} which result in non-zero value on top of the stack
                # false_values: possible values for {key} which result in zero value on top of the stack
                # if the check is not related to the field, true_values and false_values will be universal sets
                true_values, false_values = self._get_asserted(key, exit_ins_arg)

                if len(block.next) == 1:
                    # happens when bz/bnz is the last instruction in the contract and there is no default branch
                    default_branch = None
                    jump_branch = block.next[0]
                else:
                    default_branch = block.next[0]
                    jump_branch = block.next[1]

                if isinstance(block.exit_instr, BZ):
                    # jump branch is taken if the comparison is false i.e not in asserted values
                    self._path_contexts[key][jump_branch][block] = false_values
                    # default branch is taken if the comparison is true i.e in asserted values
                    if default_branch is not None:
                        self._path_contexts[key][default_branch][block] = true_values
                elif isinstance(block.exit_instr, BNZ):
                    # jump branch is taken if the comparison is true i.e in asserted values
                    self._path_contexts[key][jump_branch][block] = true_values
                    # default branch is taken if the comparison is false i.e not in asserted values
                    if default_branch is not None:
                        self._path_contexts[key][default_branch][block] = false_values

    def _update_gtxn_constraints(self, keys_with_gtxn: List[str], block: "BasicBlock") -> None:
        """Use information from txn constraints on gtxn_

        `block.transaction_context.group_indices` will contain indices the `txn` can have.

        The values of each key represent possible values for that field. Values of `GTXN_0_RekeyTo` are
        possible values for `gtxn 0 RekeyTo` i.e possible `RekeyTo` field values of transaction present at index 0.

        self._block_contexts[f"GTXN_{idx}_{key}"] stores the information collected from
        instructions `gtxn {idx} {field}`. This information represents validations performed
        on the `txn {field}` by accessing it through `gtxn {idx} {field}`.

        e.g if index of the `txn` should be 0 then `txn RekeyTo` is same as `gtxn 0 RekeyTo`.
        similary, if index of `txn` can be `0` or `1` then checking `txn RekeyTo` is equaivalent to
        checking both `gtxn 0 RekeyTo` and `gtxn 1 RekeyTo`.

        if `i` is not a possible index of `txn` for basic block `B`, then possible values for `txn {field}` when
        accessed through `gtxn {i} {field}` is Null. Because, `txn` can never have index `i` and `gtxn {i} {field}` is field
        of `txn` when index of `txn` is `i`.

        This requires that group_indices analysis is done before any other analysis.

        Args:
            keys_with_gtxn: The list of base keys whose gtxn {idx} {RekeyTo} should be tracked.
            block: The basic block

        Fill the _block_contexts[gtx_key][block] using the possible group indices of the transaction and possible
        values of the base key.
        """
        for key in keys_with_gtxn:
            for ind in range(MAX_GROUP_SIZE):
                gtx_key = get_gtxn_at_index_key(ind, key)
                if ind in self._function.transaction_context(block).group_indices:
                    # txn can have index {ind}
                    # gtxn {ind} {field} can have a value if and only if {txn} {field} can also have that value
                    self._block_contexts[gtx_key][block] = self._intersection(
                        gtx_key,
                        self._block_contexts[gtx_key][block],
                        self._block_contexts[key][block],
                    )
                else:
                    # txn cannot have index {ind}
                    self._block_contexts[gtx_key][block] = self._null_set(gtx_key)

    def _calculate_reachin(
        self, key: str, block: "BasicBlock", reachout: Dict["BasicBlock", Any]
    ) -> Any:
        if block == self._entry_block:
            # We are considering each possible value as a definition defined at the start of entry block.
            reachin_information = self._universal_set(key)
        else:
            reachin_information = self._null_set(key)

        path_context = self._path_contexts[key]
        for prev_b in prev_blocks_global(self._function, block):
            # print(f"block:", repr(block),"prev_b ", repr(prev_b), f"block: {block.__hash__()}, prev_b: {prev_b.__hash__()}")
            # print(path_context)
            # print(path_context[block])
            reachin_information = self._union(
                key,
                reachin_information,
                self._intersection(key, reachout[prev_b], path_context[block][prev_b]),
            )

        if block.is_sub_return_point:
            # this block is the return point for callsub instruction present in `block.callsub_block`
            # execution will only reach this block, if it reaches `block.callsub_block`
            reachin_information = self._intersection(
                key, reachin_information, reachout[block.callsub_block]
            )

        return reachin_information

    def _merge_information_forward(
        self,
        analysis_keys: List[str],
        block: "BasicBlock",
        global_reachout: Dict[str, Dict["BasicBlock", Any]],
    ) -> bool:
        updated = False
        for key in analysis_keys:
            # RCHout(b) = intersection(RCHin(b), PRSV(b))
            new_reachout = self._intersection(
                key,
                self._calculate_reachin(key, block, global_reachout[key]),
                self._block_contexts[key][block],
            )
            if new_reachout != global_reachout[key][block]:
                global_reachout[key][block] = new_reachout
                updated = True
        return updated

    def forward_analyis(self, analysis_keys: List[str], worklist: List["BasicBlock"]) -> None:
        """Perform forward analysis for analysis_keys and update self._block_contexts

        Args:
            analysis_keys: List of keys.
            worklist: List of basic blocks considered for the analysis. Initial worklist.
        """
        # reachout for all analysis keys. global_reachout[key] -> reachout of key.
        # global_reachout[key][block] -> reachout of block for key.
        global_reachout: Dict[str, Dict["BasicBlock", Any]] = {}
        for key in analysis_keys:
            global_reachout[key] = {}
            for b in self._function.blocks:
                global_reachout[key][b] = self._null_set(key)

        while worklist:
            b = worklist[0]
            worklist = worklist[1:]
            updated = self._merge_information_forward(analysis_keys, b, global_reachout)

            if updated:
                # if b is callsub block the information of the called subroutine and return point need to be computed.
                # called subroutine entry is already included by next_blocks_global. So, add return point here.
                return_point_block = (
                    [b.sub_return_point]
                    if b.is_callsub_block and b.sub_return_point is not None
                    else []
                )

                for bi in next_blocks_global(self._function, b) + return_point_block:
                    if bi not in worklist:
                        worklist.append(bi)

        for key in analysis_keys:
            self._block_contexts[key] = global_reachout[key]

    def _calculate_livein(
        self, key: str, block: "BasicBlock", liveout: Dict["BasicBlock", Any]
    ) -> Any:
        livein_information = self._null_set(key)

        for next_b in next_blocks_global(self._function, block):
            livein_information = self._union(key, livein_information, liveout[next_b])

        if (
            block.is_callsub_block
            and block.sub_return_point is not None
            and len(block.called_subroutine.retsub_blocks) != 0
        ):
            # this block is the `callsub block` and `block.sub_return_point` is the block that will be executed after subroutine.
            livein_information = self._intersection(
                key, livein_information, liveout[block.sub_return_point]
            )
        return livein_information

    def _merge_information_backward(
        self,
        analysis_keys: List[str],
        block: "BasicBlock",
        global_liveout: Dict[str, Dict["BasicBlock", Any]],
    ) -> bool:
        if leaf_block_global(block):  # leaf block
            return False

        updated = False
        for key in analysis_keys:
            new_liveout = self._intersection(
                key,
                self._calculate_livein(key, block, global_liveout[key]),
                self._block_contexts[key][block],
            )
            if new_liveout != global_liveout[key][block]:
                global_liveout[key][block] = new_liveout
                updated = True
        return updated

    def backward_analysis(self, analysis_keys: List[str], worklist: List["BasicBlock"]) -> None:
        """Perform backward analysis for analysis_keys and update self._block_contexts

        Args:
            analysis_keys: List of keys.
            worklist: List of basic blocks considered for the analysis. Initial worklist.
        """
        global_liveout: Dict[str, Dict["BasicBlock", Any]] = {}
        for key in analysis_keys:
            global_liveout[key] = {}
            for b in self._function.blocks:
                if leaf_block_global(b):  # leaf block
                    global_liveout[key][b] = self._block_contexts[key][b]
                else:
                    global_liveout[key][b] = self._null_set(key)

        while worklist:
            b = worklist[0]
            worklist = worklist[1:]
            updated = self._merge_information_backward(analysis_keys, b, global_liveout)

            if updated:
                callsub_block = [b.callsub_block] if b.is_sub_return_point else []
                for bi in prev_blocks_global(self._function, b) + callsub_block:
                    if bi not in worklist:
                        worklist.append(bi)

        for key in analysis_keys:
            self._block_contexts[key] = global_liveout[key]

    @staticmethod
    def _postorder(entry: "BasicBlock") -> List["BasicBlock"]:
        visited: Set["BasicBlock"] = set()
        order: List["BasicBlock"] = []

        def dfs(block: "BasicBlock") -> None:
            visited.add(block)
            for successor in block.next:
                if not successor in visited:
                    dfs(successor)
            order.append(block)

        dfs(entry)
        return order

    def run_analysis(self) -> None:  # pylint: disable=too-many-branches
        """Run analysis"""

        gtx_keys = []
        for key in self.KEYS_WITH_GTXN:
            for ind in range(MAX_GROUP_SIZE):
                gtx_keys.append(get_gtxn_at_index_key(ind, key))
                gtx_keys.append(get_absolute_index_key(ind, key))

            for offset in range(-(MAX_GROUP_SIZE - 1), MAX_GROUP_SIZE):
                if offset == 0:
                    continue
                gtx_keys.append(get_relative_index_key(offset, key))

        all_keys = self.BASE_KEYS + gtx_keys

        # step 1: initialise information
        # print(self._function.blocks)
        for block in self._function.blocks:
            self._block_level_constraints(all_keys, block)  # initialise information for all keys
            self._path_level_constraints(all_keys, block)
            for debug_key in debug_keys:
                if debug_key in self.BASE_KEYS:
                    logger_txn_ctx.debug(f"Debug key: {debug_key}")
                    logger_txn_ctx.debug(f"block: B{block.idx}:\n {block}")
                    logger_txn_ctx.debug(f"block_ctx: {self._block_contexts[debug_key][block]}")
                    logger_txn_ctx.debug("\n\n")
                    logger_txn_ctx.debug(f"path_context: {self._path_contexts[debug_key]}")
        # print(self._block_contexts)
        # The contract is represented using multiple CFG, one for each subroutine.
        # Calculate postorder for each CFG and concatenate the lists for now.
        postorder = [self._postorder(self._entry_block)]
        for subroutine in self._function.subroutines.values():
            postorder.append(self._postorder(subroutine.entry))

        # perform analysis of base keys first. Information of these base keys will be used for
        # gtxn keys. see `self._update_gtxn_constraints`
        analysis_keys = list(self.BASE_KEYS)

        worklist = []
        for l in postorder:
            worklist += l[::-1]  # Reverse postorder

        self.forward_analyis(analysis_keys, worklist)

        logger_txn_ctx.debug("After Forward:")
        for debug_key in debug_keys:
            for block in self._function.blocks:
                if debug_key in self.BASE_KEYS:
                    logger_txn_ctx.debug(f"block_id: B{block.idx}")
                    logger_txn_ctx.debug(f"block_ctx: {self._block_contexts[debug_key][block]}")

        worklist = []
        for l in postorder:
            worklist += [b for b in l if not leaf_block_global(b)]  # postorder, exclude leaf blocks

        self.backward_analysis(analysis_keys, worklist)

        logger_txn_ctx.debug("After Backward:")
        for debug_key in debug_keys:
            for block in self._function.blocks:
                if debug_key in self.BASE_KEYS:
                    logger_txn_ctx.debug(f"block_id: B{block.idx}")
                    logger_txn_ctx.debug(f"block_ctx: {self._block_contexts[debug_key][block]}")

        # update gtxn constraints using possible group indices and txn constraints.
        for block in self._function.blocks:
            self._update_gtxn_constraints(self.KEYS_WITH_GTXN, block)

        # Now perform analysis for gtx_keys
        analysis_keys = gtx_keys

        worklist = []
        for l in postorder:
            worklist += l[::-1]  # Reverse postorder
        self.forward_analyis(analysis_keys, worklist)

        worklist = []
        for l in postorder:
            worklist += [b for b in l if not leaf_block_global(b)]  # postorder, exclude leaf blocks
        self.backward_analysis(analysis_keys, worklist)

        self._store_results()



"""
TODO: Move below information to appropriate location

Literal value instructions:

- Instructions which push literal values onto the stack.
- Literal values are part of the compiled program and can be calculated using static analysis.

    intcblock, intc, intc_0, intc_1, intc_2, intc_3
    int, pushint, pushints

    bytecblock, bytec, bytec_0, bytec_1, bytec_2, bytec_3
    addr, method, bytes, pushbytes, pushbytess

    global ZeroAddress

- pushints, pushbytess push multiple literal values on to the stack.


Transaction Information:
    Block producer controlled:
        arg, arg_0, arg_1, arg_2, arg_3, args,

    Transaction signer controlled:
        txn f, gtxn t f, txna f i, gtxna t f i, gtxns f, gtxnsa f i,
        txnas f, gtxnas t f, gtxnsas f

    Information global to all transactions in the group or in the chain:
        global f

    Application/Asset Ids created by previous transactions in the group:
        gaid t, gaids,


Application state manipulation:
- Instructions to access/manipulate current application's persistent state.

    Global and Local state:
        app_global_get, app_global_put, app_global_del,
        app_local_get, app_local_put, app_local_del,

    App Boxes:
        box_create, box_put, box_del, box_replace,
        box_len, box_get, box_extract


External state information:
- Instructions to access information of external apps, assets and accounts

    app_opted_in, app_local_get_ex, app_global_get_ex, app_params_get,      # apps
    asset_holding_get, asset_params_get,                                    # assets
    balance, min_balance, acct_params_get,                                  # accounts


Block Information:
    block f


Inner transactions(Calls):
    Instructions to create a inner transaction:
        itxn_begin, itxn_field, itxn_next, itxn_submit,

    Instructions to access information of recently submitted inner transaction:
        itxn f, itxna f i, gitxn t f, gitxna t f i, itxnas f, gitxnas t f,


Stack Manipulation Opcodes:

- Instructions which change the order of elements in the stack independent of the actual values of the elements.
- could be emulated during stack emulation

    bury n, popn n, dupn n, pop, dup,
    dup2, dig n, swap, cover n, uncover n,

    select, # select depends on the runtime information


Scratch Space:
    Instructions to manipulate scratch space:
        load, store, loads, stores,

    Instructions to access scratch space of previous transactions in the group:
        gload t i, gloads i, gloadss


Control Flow:
    (Might) Stop the execution:
        err, return, assert,

    Branch, Jump around the code:
        b target
        bnz target, bz target, switch, match,

        callsub, retsub,


Arthimetic Operations:
    +, -, /, *, %, sqrt, exp, shl, shr,
    addw, mulw, divw, divmodw, expw,        # wide math
    b+, b-, b/, b*, b%, bsqrt,              # byte (512-bits)


Bitwise Operations:
    |, &, ^, ~,
    b|, b&, b^, b~,         # byte (512-bits)


Connectors:
    &&
    ||

Comparison
    <, >, <=, >=, ==, !=, !
    b<, b>, b<=, b>=, b==, b!=


Int/Byte Manipulation:
    bzero,
    concat,
    len, bitlen,
    itob, btoi,
    getbit, setbit, getbyte, setbyte,
    substring s e, substring3,
    extract, extract3, extract_uint16, extract_uint32, extract_uint64,
    replace2, replace3,
    base64_decode, json_ref,


Cryptographic/Hash:
    sha256, keccak256, sha512_256, sha3_256,
    ed25519verify, ed25519verify_bare,
    ecdsa_verify, ecdsa_pk_decompress, ecdsa_pk_recover
    vrf_verify,


Logging Operations:
    log


Unclear/recent(v8) instructions:
    proto, frame_dig, frame_bury,
"""

from typing import List, Union, Dict, Type, Tuple
from functools import lru_cache


from tealer.teal.basic_blocks import BasicBlock
from tealer.teal.instructions.instructions import (
    Instruction,
)


class UnknownStackValue:
    """Represent a unknown stack value."""

    def __str__(self) -> str:
        return "UnknownStackValue()"

    def __repr__(self) -> str:
        return "UnknownStackValue()"


class KnownStackValue:
    """Represent values which are known during the local stack emulation.

    A stack value is represented using the instruction, arguments and the position.

    instruction: Every stack value is a result of executing a instruction. instruction
        property contains the reference to that instruction's object.
    args: A instruction consumes values from the stack. `args` represents the list of stack values
        which are consumed by `self.instruction` to produce the result(s).
    ins_out_values_index: ins_out_values_index stores the index inside the list of instruction
    out values. When a instruction pushes multiple out values, every out value will have same
    `instruction` and `args`. ins_out_values_index allows representing out values uniquely.


    e.g

    int 10
    int 11
    mulw

    Stack:

        |               |
         ---------------
        |               |
         ---------------

    `int 10` -> pops 0 values and pushes 1 value

        |               |
         ---------------
        |  A = Int(10)  |
         ---------------

    `int 11` -> pops 0 values and pushes 1 value

        |               |
         ---------------
        |  B = Int(11)  |
         ---------------
        |  A = Int(10)  |
         ---------------

    mulw -> pops 2 values and pushes 2 values.

        |               |
         ---------------
        |       Y       |
         ---------------
        |       X       |
         ---------------

    A = KnownStackValue(ins = Int(10), args = [], ins_out_values_index = 0)
    B = KnownStackValue(ins = Int(11), args = [], ins_out_values_index = 0)

    mulw pops `a` and `b` and pushes `X` and `Y`.
    mulw_out_values = mulw(a, b)
    X = mulw_out_values[0]
    Y = mulw_out_values[1]

    X == KnownStackValue(ins = Mulw(), args = [A, B], ins_out_values_index = 0)
    Y == KnownStackValue(ins = Mulw(), args = [A, B], ins_out_values_index = 1)
    """

    def __init__(self, ins: Instruction, args: List, ins_out_values_index: int = 0) -> None:
        self._ins = ins
        self._args: List[Union[KnownStackValue, UnknownStackValue]] = args
        self._ins_out_values_index = ins_out_values_index

    @property
    def instruction(self) -> Instruction:
        return self._ins

    @property
    def args(self) -> List[Union["KnownStackValue", UnknownStackValue]]:
        """List of stack values consumed by self.instruction.

        self.args[0] is the last popped(first pushed) value from the stack.
        self.args[-1] is the first popped value from the stack:
            self.args[-1] is on top of the stack right before the execution
            of self.instruction.

        Returns:
            Arguments of the current instruction that are taken from the stack.
        """
        return self._args

    @property
    def ins_out_values_index(self) -> int:
        """index into the list of out values pushed by self.instruction.

        An instruction can push multiple stack values. In order to differentiate
        between multiple values generated by the execution of the same instruction,
        the ins_out_values_index is used.

        The stack values pushed first by the instruction will have index 0. The stack value
        pushed next will have index 1 and so on.

        stack:
                 ---- ----
                |    |
                 ---- ----
        int 10
                 ---- ----
                | 10 |
                 ---- ----
        int 11
                 ---- ---- ----
                | 10 | 11 |
                 ---- ---- ----
        mulw
                 ---- ---- ----
                |  X |  Y |
                 ---- ---- ----

        (X, Y) = mulw(10, 11)

        mulw_in_values = Int(10), Int(11)
        mulw_out_values = [X, Y]

        X.ins_out_values_index == 0
        Y.ins_out_values_index == 1

        X.instruction == Y.instruction == Mulw()
        X.args == Y.args = [Int(10), Int(11)]

        Returns:
            index of this value in the list of output values of self.instruction.
        """
        return self._ins_out_values_index

    def __str__(self) -> str:
        # args = ", ".join(str(i) for i in self._args)
        # return f"{self._ins.__class__.__name__}(<{args}>)"
        args = "[" + ", ".join(str(i) for i in self.args) + "]"
        # return f"KnownStackValue(ins = {str(self._ins.__class__.__name__)}, args = {args})"
        return f"KnownStackValue(ins = {repr(self._ins)}, args = {args})"

    def __repr__(self) -> str:
        return str(self)


StackValue = Union[KnownStackValue, UnknownStackValue]


class Stack:
    """Represent a stack.

    Elements are stored in a list. Top of the stack is at the end of the List.

    Stack is assumed to have infinite UnknownStackValue below the current pointer.
    """

    def __init__(self) -> None:
        self._values: List[StackValue] = []

    def push_n_values(self, values: List[StackValue]) -> None:
        self._values.extend(values)

    def pop_n_values(self, count: int) -> List[StackValue]:
        """Pop elements from the stack.

        Args:
            count: Number of elements to pop.

        Returns:
            list of popped values. First popped value is at the end of the list.
        """
        if count == 0:
            return []
        if len(self._values) >= count:
            vals = self._values[-count:]
            self._values = self._values[:-count]
            return vals
        remaining_elements: List[StackValue] = [
            UnknownStackValue() for _ in range(count - len(self._values))
        ]
        vals = remaining_elements + self._values
        self._values = []
        return vals

    def __str__(self) -> str:
        return "[" + ",".join([str(value) for value in self._values]) + "]"

    def __repr__(self) -> str:
        return str(self)


@lru_cache(maxsize=None)
def construct_stack_ast(bb: BasicBlock) -> Dict[Instruction, KnownStackValue]:
    r"""Emulates instructions in the basic block and return values(AST) produced by each instruction.

    Because we do not know the values of the stack before executing this basic block, stack is assumed
    to have infinite number of UnknownStackValue(s).

    For each instruction in the basic block, this function calculates the AST consisting of instructions.

    int 1
    int 2
    ==              // <1>
    addr "..."
    txn RekeyTo
    !=              // <2>
    &&
    byte "X"
    byte "Y"
    ==              // <3>
    &&

    Tree:
                        And
                      /     \\
                    And       Eq     <- <3>
                  /    \    /     \\
        <1> ->  Eq     Neq  Byte   Byte
              /   \   /   \\
            Int  Int Addr  Txn RekeyTo

    int 1
    &&
    assert
    Returned dict:
        {
            Int(1): KnownStackValue(ins = Int(1), args = []),
            And(): KnownStackValue(
                    ins = And(),
                    args = [
                    UnknownStackValue,
                    KnownStackValue(ins = Int(1), args = []),
                ]),
            Assert(): KnownStackValue(
                    ins = Assert(),
                    args = [
                        KnownStackValue(
                            ins = And(),
                            args = [
                            UnknownStackValue,
                            KnownStackValue(ins = Int(1), args = []),
                        ]),
                ]),
        }

    Args:
        bb: A basic block of the CFG.

    Returns:
        A dict with each instruction of the bb as key. The values are the stack value representation
        of that instruction. The value should be used to access the inputs of the instruction. These
        inputs only includes stack arguments not the immediate arguments.
    """
    stack = Stack()
    ins_stack_value = {}
    for ins in bb.instructions:
        # TODO: emulate other stack manipulating instructions if necessary:
        #   swap, dup, dup2, dupn n, bury n, dig n, cover n, uncover n,
        # - Replace int like instructions with int values. Same for byte like instructions
        # - Only track necessary instructions values.
        ins_in_values = stack.pop_n_values(ins.stack_pop_size)  # pops `pop_size` number of values
        ins_out_values: List[StackValue] = []
        for i in range(ins.stack_push_size):
            ins_out_values.append(KnownStackValue(ins, ins_in_values, i))
        stack.push_n_values(ins_out_values)

        # store the instruction reference and its arguments(input values).
        # we only care about the instruction input values for now.
        ins_stack_value[ins] = KnownStackValue(ins, ins_in_values)
    return ins_stack_value


def _flatten_ast(root: StackValue, node_ins: Type[Instruction]) -> List[StackValue]:
    r"""Return all the leaf values given the root of the instruction AST.

    node_ins instruction is one of `And`, `Or`.
    `And` and `Or` consume two values. These two values are the children of the node.

    Inner nodes are values whose `instruction` is a instance of `node_ins`.
    Leaf values are values whose `instruction` is not `node_ins`.

      &&
    /    \
   ==     &&
        /    \
       &&     ||
      /  \   /  \
     ==  >  !   !=
    /\   /\

    All the nodes in the tree are represented by `StackValue`. A StackValue can be unknown or known.
    unknown stack values are considered to be leaf values and are stored.
    A known stack value contains the instruction and arguments consumed by that instruction. These
    arguments are also stack values and are treated as children.

    In above example, if `node_ins` is `And` then values with instruction, `==, >, ||`, are all treated
    as leaf blocks and are NOT traversed further.

    int 1
    int 2
    ==
    addr "..."
    txn RekeyTo
    !=
    &&
    byte "X"
    byte "Y"
    ==
    &&

    Tree:
                        And
                      /     \
                    And       Eq
                  /    \    /     \
                Eq     Neq  Byte   Byte
              /   \   /    \
            Int  Int Addr  Txn RekeyTo

    node_ins = And

    ^^Only And instruction values are considered as nodes. remaining values are treated as leaves
    and are not traversed.

    After traversing, this functions returns the leaf values:
        [
            Eq(Int(<>), Int(<>)),
            Neq(Addr(<>), Txn(<>)),
            Eq(Byte(<>), Byte(<>))
        ]

    This function essentially flattens the given tree:
        And(And(<1>, <2>), And(And(<3>, <4>), And(<5>, <6>)))
        node_ins = And
        => returns [<1>, <2>, <3>, <4>, <5>, <6>]

        Or(Or(Or(<1>, <2>), Or(<3>, <4>)), Or(<5>, <6>))
        node_ins = Or
        => returns [<1>, <2>, <3>, <4>, <5>, <6>]

    Args:
        root: root stack value of the instruction tree
        node_ins: node_ins is used to differentiate to identify leaf nodes.

    Returns:
        List of leaf nodes. Leaf nodes are stack values whose instruction is not :node_ins: and
        All stack values in the path from the root to the leaf node have :node_ins: instruction.
    """
    if isinstance(root, UnknownStackValue):
        return [root]
    if not isinstance(root.instruction, node_ins):
        # is not node_ins stack value. so a leaf
        return [root]
    left, right = root.args[0], root.args[1]
    return _flatten_ast(left, node_ins) + _flatten_ast(right, node_ins)


@lru_cache(maxsize=None)
def compute_equations(
    root: KnownStackValue, node_ins: Type[Instruction]
) -> Tuple[List[KnownStackValue], bool]:
    """Compute equations from And or Or expression.

    Args:
        root: Root of the instruction AST
        node_ins: The instruction whose result is being flattend. see documentation for _flatten_ast
            function above.

    Returns:
        equations: List of values on which, the result of **root** depends.

            And(And(<1>, <2>), And(And(<3>, <4>), And(<5>, <6>)))
            node_ins = And
            => returns [<1>, <2>, <3>, <4>, <5>, <6>]

            Or(Or(Or(<1>, <2>), Or(<3>, <4>)), Or(<5>, <6>))
            node_ins = Or
            => returns [<1>, <2>, <3>, <4>, <5>, <6>]

        has_unknown_value: True if the result of **root** depends on a unknown value.
    """
    equations = _flatten_ast(root, node_ins)
    has_unkown_value = False
    known_equations = []
    for eq in equations:
        if isinstance(eq, UnknownStackValue):
            has_unkown_value = True
        else:
            known_equations.append(eq)
    return known_equations, has_unkown_value


def get_stack_value_for_ins(ins: "Instruction") -> KnownStackValue:
    return construct_stack_ast(ins.bb)[ins]



"""Entry point of **Tealer** when executed as a script.

Tealer is a static analysis tool for Algorand Smart Contracts written
in TEAL. Tealer takes the source(teal) code of the smart contract as
input, parses and runs detectors, printers on the parsed code.

Detectors are analyzers which go through the parsed code and look for
certain patterns in the program which are of interest to anyone invested
in the security of the contract.

Printers are information summarizers. Printers make finding and evaluating
interesting information related to the contract easier. There's always abundant
information that can be extracted from a smart contract, manually going
through a smart contract for a every tiny bit of information is hard, boringgg
and error prone. Printers make this process of finding information easier and
colorful(will be after adding colors and ofcourse emojis, who doesn't like them <3).

Tealer comes with builtin detectors for finding common security issues and builtin
printers for printing information interesting in general. Additional to that,
tealer allows adding user defined detectors, printers in the form of plugins.
With the plugin support, tealer makes adding new detectors, printers easier and
allows to get away with the minimum amount of work.

Once installed, Tealer adds a command ``tealer`` to the system which will invoke
the ``main()`` function when executed. Tealer comes with few command line
arguments which help users in selecting and running specific detectors, printers
based on the need.

The normal execution flow of tealer is:

* Select the detectors, printers to run on the contract.
    * Collect the list of detectors, printers that come with tealer.
    * Find the plugins installed in the system and collect their detectors, printers.
    * Choose the detectors, printers based on the user command line arguments.
* Read the contract teal source code.
* Parse the source code and generate Teal object.
    * Parse each line of the source as a teal instruction.
    * Construct basic blocks using branch instructions and other rules.
    * Construct Control Flow Graph(CFG) of the contract.
    * Detect basic info like version, mode and create a Teal object.
* Run the selected detectors, printers on the contract.
    * Detectors, Printers are registered, run on the contract using Teal object API.
    * The results are displayed in json format or normally based on command line argument.

The following are few of the command line arguments a user can use to get the best out of tealer:

To know available detectors and printers, both tealer builtins as well as installed plugins:

--list-detectors - prints a pretty table listing the available detectors.
--list-printers - prints a pretty table listing the available printers.

To choose detectors to run from the available list:

--detect - takes in comma separated list of detectors to run. runs all of them if not given.
--exclude - takes in comma separated list of detectors to exclude from running detectors list.
--exclude-stateless - excludes detectors that are supposed to run on stateless smart contracts.
--exclude-stateful - excludes detectors that are supposed to run on stateful smart contracts.

To choose printers to run from the available list:

--print - takes in comma separated list of printers to run. doesn't run any if not given.

"""

import argparse
import json
import logging
import os
import re
import sys
from pathlib import Path
from typing import List, Any, Type, Tuple, TYPE_CHECKING, Optional, Union, Sequence

from pkg_resources import require  # type: ignore

from tealer.detectors.abstract_detector import AbstractDetector, DetectorType
from tealer.exceptions import TealerException
from tealer.printers.abstract_printer import AbstractPrinter
from tealer.utils.algoexplorer import (
    get_application_using_app_id,
    logic_sig_from_contract_account,
    logic_sig_from_txn_id,
)
from tealer.utils.command_line.command_output import (
    output_detectors,
    output_printers,
    output_to_markdown,
    output_wiki,
)
from tealer.utils.command_line.common import (
    get_detectors_and_printers,
    validate_command_line_options,
    # handle_detect,
    init_tealer_from_single_contract,
    init_tealer_from_config,
)
from tealer.utils.command_line.group_config import (
    read_config_from_file,
)
from tealer.utils.output import ROOT_OUTPUT_DIRECTORY, ExecutionPaths
from tealer.utils.regex.regex import run_regex
from tealer.utils.teal_enums import ExecutionMode

if TYPE_CHECKING:
    from tealer.teal.teal import Teal
    from tealer.utils.output import ListOutput


# from slither: slither/__main__.py
def choose_detectors(
    args: argparse.Namespace,
    all_detector_classes: List[Type[AbstractDetector]],
    teal: Optional["Teal"] = None,
) -> List[Type[AbstractDetector]]:
    """Select detectors from available list based on command line arguments.

    Detectors are selected using the values of command line arguments
    ``--detect``, ``--exclude``, ``--exclude-stateless``, ``--exclude-stateful``.
    If the default detectors are run, exclude the stateless detector for stateful app, and stateful detectors
    for stateless code. The detectors can still be executed if they are added manually through --detect

    Args:
        args: Namespace object representing the command line arguments selected
            by the user.
        all_detector_classes: list of all available detectors.
        teal: The Teal contract.

    Returns:
        list of chosen detectors from the available list using the tealer
        command line arguments.

    Raises:
        TealerException: raises exception if name of detector selected using the
            command line argument is not present in the given availble list of
            detectors.
    """

    detectors_to_run = []
    detectors = {d.NAME: d for d in all_detector_classes}

    if args.detectors_to_run is None:
        detectors_to_run = all_detector_classes
        # IF there is no detectors provided:
        # - Stateful: run everything expect the stateless detectors
        # - Stateless: run only stateless and stateless & stateful
        if teal is not None and teal.mode == ExecutionMode.STATEFUL:
            detectors_to_run = [d for d in detectors_to_run if d.TYPE != DetectorType.STATELESS]
        if teal is not None and teal.mode == ExecutionMode.STATELESS:
            detectors_to_run = [
                d
                for d in detectors_to_run
                if d.TYPE in [DetectorType.STATELESS, DetectorType.STATELESS_AND_STATEFULL]
            ]
    else:
        for detector in args.detectors_to_run.split(","):
            detector_name = detector.strip()
            if detector_name in detectors:
                detectors_to_run.append(detectors[detector_name])
            else:
                raise TealerException(f"Error: {detector_name} is not a detector")

    if args.detectors_to_exclude:
        detectors_to_run = [d for d in detectors_to_run if d.NAME not in args.detectors_to_exclude]

    if args.exclude_stateful:
        detectors_to_run = [d for d in detectors_to_run if d.TYPE != DetectorType.STATEFULL]

    if args.exclude_stateless:
        detectors_to_run = [d for d in detectors_to_run if d.TYPE != DetectorType.STATELESS]

    return detectors_to_run


# from slither: slither/__main__.py
def choose_printers(
    args: argparse.Namespace, all_printer_classes: List[Type[AbstractPrinter]]
) -> List[Type[AbstractPrinter]]:
    """Select printers from available list based on command line arguments.

    printers are selected using the value of command line argument
    ``--print`` which is comma separated list of printer names to run.

    Args:
        args: Namespace object representing the command line arguments selected
            by the user.
        all_printer_classes: list of all available printers.

    Returns:
        list of chosen printers from the available list based on tealer
        command line arguments.

    Raises:
        TealerException: raises exception if name of printer selected using the
            command line argument is not present in the given availble list of
            printers.
    """

    if args.printers_to_run is None:
        return []

    printers = {printer.NAME: printer for printer in all_printer_classes}
    printers_to_run = []
    for printer in args.printers_to_run.split(","):
        if printer in printers:
            printers_to_run.append(printers[printer])
        else:
            raise TealerException(f"{printer} is not a printer")
    return printers_to_run


def generic_options(parser: argparse.ArgumentParser) -> None:
    """
    Add the generic cli flags options

    Args:
        parser: the parser to update
    """

    parser.add_argument(
        "--contracts",
        help="List of files/directories to search for .teal files",
        action="store",
        nargs="+",
    )

    parser.add_argument(
        "--group-config",
        help="The group configuration file",
        action="store",
    )

    parser.add_argument(
        "--network",
        help='Algorand network to fetch the contract from, ("mainnet" or "testnet"). defaults to "mainnet".',
        action="store",
        default="mainnet",
    )


def parse_args(
    detector_classes: List[Type[AbstractDetector]], printer_classes: List[Type[AbstractPrinter]]
) -> argparse.Namespace:
    """parse command line arguments of tealer.

    Args:
        detector_classes: list of detectors available to tealer. Used to display
            available detectors in tealer command help message.
        printer_classes: list of printers available to tealer. Used to display
            available printers in tealer command help message.

    Returns:
        Namespace object representing the parsed command line arguments.
    """

    parser = argparse.ArgumentParser(
        description="TealAnalyzer",
        usage="tealer [detect | print | regex ] --contracts ...",
    )

    subparsers = parser.add_subparsers(dest="subcommand")

    group_detector = subparsers.add_parser("detect", help="Run the detectors")
    generic_options(group_detector)

    available_printers = ", ".join(p.NAME for p in printer_classes)
    group_printer = subparsers.add_parser("print", help="Use a printer")
    group_printer.add_argument(
        "printers_to_run",
        action="store",
        help=f"Comma-separated list of printers to use, defaults to None. Available printers: {available_printers}",
    )
    generic_options(group_printer)

    group_regex = subparsers.add_parser("regex", help="Use the regex engine")
    group_regex.add_argument("regex_file", action="store", help="Regular expression file")
    generic_options(group_regex)

    group_misc = parser.add_argument_group("Additional options")

    parser.add_argument(
        "--version",
        help="displays the current version",
        version=require("tealer")[0].version,
        action="version",
    )

    available_detectors = ", ".join(d.NAME for d in detector_classes)

    group_detector.add_argument(
        "--detectors",
        help="Comma-separated list of detectors, defaults to all, "
        f"available detectors: {available_detectors}",
        action="store",
        dest="detectors_to_run",
        default=available_detectors,
    )

    group_detector.add_argument(
        "--exclude",
        help="Comma-separated list of detectors that should be excluded.",
        action="store",
        dest="detectors_to_exclude",
        nargs="+",  # Allows multiple arguments
        default=None,
    )

    group_detector.add_argument(
        "--exclude-stateless",
        help="Exclude detectors of stateless type",
        action="store_true",
        default=False,
    )

    group_detector.add_argument(
        "--exclude-stateful",
        help="Exclude detectors of stateful type",
        action="store_true",
        default=False,
    )

    group_detector.add_argument(
        "--filter-paths",
        help="Excludes execution paths matching the regex from detector's output.",
        action="store",
        dest="filter_paths",
        default=None,
    )

    group_detector.add_argument(
        "--list-detectors",
        help="List available detectors",
        action=ListDetectors,
        nargs=0,
        default=False,
    )

    group_misc.add_argument(
        "--json",
        help='Export the results as a JSON file ("--json -" to export to stdout)',
        action="store",
        default=None,
    )

    group_printer.add_argument(
        "--list-printers",
        help="List available printers",
        action=ListPrinters,
        nargs=0,
        default=False,
    )

    parser.add_argument("--debug", help=argparse.SUPPRESS, action="store_true", default=False)
    parser.add_argument("--markdown", help=argparse.SUPPRESS, action=OutputMarkdown, default=False)

    parser.add_argument(
        "--wiki-detectors", help=argparse.SUPPRESS, action=OutputWiki, default=False
    )

    if len(sys.argv) == 1:
        parser.print_help(sys.stderr)
        sys.exit(1)

    args = parser.parse_args()

    return args


# from slither: slither/__main__.py
class ListDetectors(argparse.Action):  # pylint: disable=too-few-public-methods
    """Argparse Action class to display available detectors.

    This Action will be invoked if ``--list-detectors`` command line
    argument is selected. Upon invocation, prints a table listing the available
    detectors along with their name, description, type, impact and confidence.
    """

    def __call__(
        self, parser: argparse.ArgumentParser, *args: Any, **kwargs: Any
    ) -> None:  # pylint: disable=signature-differs
        detectors, _ = get_detectors_and_printers()
        output_detectors(detectors)
        parser.exit()


# from slither: slither/__main__.py
class ListPrinters(argparse.Action):  # pylint: disable=too-few-public-methods
    """Argparse Action class to display available printers.

    This Action will be invoked if ``--list-printers`` command line
    argument is selected. Upon invocation, prints a table listing the available
    printers along with their name and description.
    """

    def __call__(
        self, parser: argparse.ArgumentParser, *args: Any, **kwargs: Any
    ) -> None:  # pylint: disable=signature-differs
        _, printers = get_detectors_and_printers()
        output_printers(printers)
        parser.exit()


class OutputMarkdown(argparse.Action):  # pylint: disable=too-few-public-methods
    def __call__(
        self,
        parser: Any,
        args: Any,
        values: Optional[Union[str, Sequence[Any]]],
        option_string: Any = None,
    ) -> None:
        detectors, printers = get_detectors_and_printers()
        assert isinstance(values, str)
        output_to_markdown(detectors, printers, values)
        parser.exit()


class OutputWiki(argparse.Action):  # pylint: disable=too-few-public-methods
    def __call__(
        self,
        parser: Any,
        args: Any,
        values: Optional[Union[str, Sequence[Any]]],
        option_string: Any = None,
    ) -> None:
        detectors, _ = get_detectors_and_printers()
        assert isinstance(values, str)
        output_wiki(detectors, values)
        parser.exit()


def handle_output(
    args: argparse.Namespace,
    detector_results: List["ListOutput"],
    teal: "Teal",
    error: Optional[str],
) -> None:
    """Util function to output tealer results.

    Tealer supports displaying the results in json format or raw format.
    Format is decided on whether ``--json`` command line argument is set or not.

    Args:
        args: Namespace object representing the command line arguments selected
            by the user.
        detector_results: results of running the selected detectors.
        teal: The contract.
        error: Error string if any detector or printer resulted in an error.
    """
    expanded_detector_results: "ListOutput" = []
    for result in detector_results:
        expanded_detector_results.extend(result)

    output_directory = ROOT_OUTPUT_DIRECTORY / Path(teal.contract_name)
    os.makedirs(output_directory, exist_ok=True)
    if args.json is None:

        if error is not None:
            print(f"Error: {error}")
            sys.exit(-1)

        detectors_with_0_results: List["AbstractDetector"] = []
        for output in expanded_detector_results:
            if not output.generate_output(output_directory):
                detectors_with_0_results.append(output.detector)
        if detectors_with_0_results:
            detectors_with_0_results_str = ", ".join(d.NAME for d in detectors_with_0_results)
            print(f"\n 0 results found for {detectors_with_0_results_str}.")
    else:
        json_results = [output.to_json() for output in expanded_detector_results]

        json_output = {
            "success": error is not None,
            "error": error,
            "result": json_results,
        }

        if args.json == "-":
            print(json.dumps(json_output, indent=2))
        else:
            filename = output_directory / Path(args.json)
            print(f"json output is written to {filename}")
            with open(filename, "w", encoding="utf-8") as f:
                f.write(json.dumps(json_output, indent=2))


def fetch_contract(args: argparse.Namespace) -> Tuple[str, str]:
    program: str = args.contracts[0]
    network: str = args.network
    b32_regex = "[A-Z2-7]+"
    if program.isdigit():
        # is a number so a app id
        print(f'Fetching application using id "{program}"')
        return get_application_using_app_id(network, int(program)), f"app-{program}"
    if len(program) == 52 and re.fullmatch(b32_regex, program) is not None:
        # is a txn id: base32 encoded. length after encoding == 52
        print(f'Fetching logic-sig contract that signed the transaction "{program}"')
        return logic_sig_from_txn_id(network, program), f"txn-{program[:8].lower()}"
    if len(program) == 58 and re.fullmatch(b32_regex, program) is not None:
        # is a address. base32 encoded. length after encoding == 58
        print(f'Fetching logic-sig of contract account "{program}"')
        return logic_sig_from_contract_account(network, program), f"lsig-acc-{program[:8].lower()}"
    # file path
    print(f'Reading contract from file: "{program}"')
    try:
        contract_name = program[: -len(".teal")] if program.endswith(".teal") else program
        contract_name = contract_name.lower()
        with open(program, encoding="utf-8") as f:
            return f.read(), contract_name
    except FileNotFoundError as e:
        raise TealerException from e


# pylint: disable=too-many-locals,too-many-branches
def main() -> None:
    """Entry point of the tealer tool.

    This function is called when tealer command is executed or tealer is
    executed as a script. This function directs execution flow of the tool
    based on the command line arguments given by the user.
    """

    detector_classes, printer_classes = get_detectors_and_printers()
    args = parse_args(detector_classes, printer_classes)
    validate_command_line_options(args)

    default_log = logging.INFO if not args.debug else logging.DEBUG
    for (logger_name, logger_level) in [
        ("Detectors", default_log),
        ("TransactionCtxAnalysis", default_log),
        ("Parsing", default_log),
        ("Tealer", default_log),
    ]:
        logger = logging.getLogger(logger_name)
        logger.setLevel(logger_level)

    results_detectors: List["ListOutput"] = []
    error = None

    if (
        args.subcommand == "detect"
        and args.detectors_to_run is not None
        and args.group_config is not None
    ):
        # handle this case specially for now.
        # handle_detect(args)
        # sys.exit(1)
        group_config = read_config_from_file(Path(args.group_config))
        tealer = init_tealer_from_config(group_config)
        detector_classes = choose_detectors(args, detector_classes)
        for detector in detector_classes:
            tealer.register_detector(detector)
        group_results = tealer.run_detectors()[0]
        for output in group_results:
            # dest is ignored
            output.generate_output(Path("."))
        sys.exit(1)

    try:  # pylint: disable=too-many-nested-blocks
        contract_source, contract_name = fetch_contract(args)
        tealer = init_tealer_from_single_contract(contract_source, contract_name)

        # TODO: handle this as a subcommand instead of a flag
        if args.subcommand == "regex":
            default_path = Path("regex_result.dot")
            if len(tealer.contracts) != 1:
                print("Regex works only for single contract")
                return
            run_regex(tealer.contracts[contract_name], args.regex_file, default_path)
            return

        # TODO: decide on default classification for detectors in group transaction context.

        if args.subcommand == "print":
            printer_classes = choose_printers(args, printer_classes)

            for printer_cls in printer_classes:
                tealer.register_printer(printer_cls)

            _results_printers = tealer.run_printers()
            return

        if args.subcommand == "detect":
            detector_classes = choose_detectors(
                args, detector_classes, tealer.contracts[contract_name]
            )
            for detector_cls in detector_classes:
                tealer.register_detector(detector_cls)

            results_detectors = tealer.run_detectors()

            if args.filter_paths is not None:
                for detector_result in results_detectors:
                    if isinstance(detector_result, ExecutionPaths):
                        detector_result.filter_paths(args.filter_paths)
                    else:
                        for result in detector_result:
                            result.filter_paths(args.filter_paths)
            # Current does not return
            # This is to call handle_output, which also catch the error
            # This should be refactored (TODO)

        else:
            print(f"{args.subcommand} not implemented")
            return

    except TealerException as e:
        error = str(e)

    # TODO: refactor this logic
    if error or args.subcommand == "detect":
        handle_output(args, results_detectors, tealer.contracts[contract_name], error)


if __name__ == "__main__":
    main()



"""Static analyzer for Algorand Smart Contracts.

Tealer comes with the utilities to parse and represent
smart contracts written in TEAL. Tealer also contains
detectors that analyze the TEAL contracts and printers
that print interesting information extracted from the
contract.

Modules:
    detectors: Module for tealer detectors.

    printers: Module for tealer printers

    teal: Contains functions, classes for parsing and representing teal contracts.

    utils: Common util functions used by other packages in Tealer.

    exceptions: Defines exceptions classes for representing exceptions in tealer.
"""



"""Printer to print human readable summary of the contract.

Summary includes :
* Contract version.
* Type of the contract as detected by tealer.
* Number of basic blocks in the contract.
* Number of instructions in the contract.
* Number of subroutines defined in the contract.
* Name of each subroutine defined in the contract.
* Number of issues found by detectors each class of impact.
* Check whether contract code is complex or not.

"""

from typing import TYPE_CHECKING

from tealer.printers.abstract_printer import AbstractPrinter
from tealer.utils.code_complexity import compute_cyclomatic_complexity

if TYPE_CHECKING:
    from pathlib import Path


# pylint: disable=too-few-public-methods
class PrinterHumanSummary(AbstractPrinter):
    """Printer to print summary of the contract"""

    NAME = "human-summary"
    HELP = "Print a human-readable summary of the contract"
    WIKI_URL = "https://github.com/crytic/tealer/wiki/Printer-documentation#human-summary"

    def _is_complex_code(self) -> str:
        """Check whether contract code is complex or not.

        Returns:
            (str): returns "Yes" if cyclomatic complexity of the contract
            is greater than 7 or else returns "No".
        """

        is_complex = False
        if compute_cyclomatic_complexity(self.teal.bbs) > 7:
            is_complex = True

        result = "Yes" if is_complex else "No"
        return result

        # pylint: disable=too-many-locals
        # TODO: Decide on how printers should behave in group context
        # def _get_detectors_result(self) -> Tuple[List, int, int, int, int, int]:
        #     """Return dectector results and number of issues found in the contract for each `impact` class.

        #     Returns:
        #         (Tuple[List, int, int, int, int, int]): returns list of results found by all the detectors and
        #         number of issues found by detectors of impact optimization, informational, low, medium, high in
        #         the same order.

        #     """
        #     from tealer.utils.command_line.common import (  # pylint: disable=import-outside-toplevel
        #         get_detectors_and_printers,
        #     )

        #     detector_classes, _ = get_detectors_and_printers()
        #     detectors = [d(self.teal) for d in detector_classes]

        #     checks_optimization = [
        #         d for d in detectors if d.IMPACT == DetectorClassification.OPTIMIZATION
        #     ]
        #     checks_informational = [
        #         d for d in detectors if d.IMPACT == DetectorClassification.INFORMATIONAL
        #     ]
        #     checks_low = [d for d in detectors if d.IMPACT == DetectorClassification.LOW]
        #     checks_medium = [d for d in detectors if d.IMPACT == DetectorClassification.MEDIUM]
        #     checks_high = [d for d in detectors if d.IMPACT == DetectorClassification.HIGH]

        #     issues_optimization = [c.detect().paths for c in checks_optimization]
        #     issues_optimization = [c for c in issues_optimization if c]

        #     issues_informational = [c.detect().paths for c in checks_informational]
        #     issues_informational = [c for c in issues_informational if c]

        #     issues_low = [c.detect().paths for c in checks_low]
        #     issues_low = [c for c in issues_low if c]

        #     issues_medium = [c.detect().paths for c in checks_medium]
        #     issues_medium = [c for c in issues_medium if c]

        #     issues_high = [c.detect().paths for c in checks_high]
        #     issues_high = [c for c in issues_high if c]

        #     all_results = (
        #         issues_optimization + issues_informational + issues_low + issues_medium + issues_high
        #     )

        #     return (
        #         all_results,
        #         len(issues_optimization),
        #         len(issues_informational),
        #         len(issues_low),
        #         len(issues_medium),
        #         len(issues_high),
        #     )

        # def get_detectors_result(self) -> str:
        # """Return textual summary of number of issues found by detectors.

        # Textual summary contains number of issues found in the contract
        # for each class of impact(severity).

        # Returns:
        #     (str): textual summary of number of issues found by detectors.
        # """

        # (
        #     _,
        #     optimization,
        #     informational,
        #     low,
        #     medium,
        #     high,
        # ) = self._get_detectors_result()

        # txt = f"Number of optimization issues: {optimization}\n"
        # txt += f"Number of informational issues: {informational}\n"
        # txt += f"Number of low issues: {low}\n"
        # txt += f"Number of medium issues: {medium}\n"
        # txt += f"Number of high issues: {high}\n"

        # return txt

    def print(self) -> None:
        """Print summary of the contract to stdout.

        Currently displays version, mode, number of basic blocks, instructions, subroutines,
        Names of subroutines, number of results found by detectors for each level of impact.
        """

        teal = self.teal

        txt = "\n"
        txt += f"Program version: {teal.version}\n"
        txt += f"Mode: {str(teal.mode)}\n"
        txt += f"Number of basic blocks: {len(teal.bbs)}\n"
        txt += f"Number of instructions: {len(teal.instructions)}\n"

        txt += f"Number of subroutines: {len(teal.subroutines.items())}\n"
        txt += "Subroutines:\n"
        for sub_name in teal.subroutines:
            txt += f"\t{sub_name}\n"
            block_ids = ", ".join(repr(bi) for bi in teal.subroutines[sub_name].blocks)
            txt += f'\t\t"{block_ids}"\n'
        # txt += f"is_complex: {self._is_complex_code()}\n" # update printer after deciding on its semantics
        print(txt)
        # print(self.get_detectors_result())



from typing import TYPE_CHECKING, List, Tuple, Optional
from dataclasses import dataclass

from tealer.analyses.dataflow.transaction_context.generic import DataflowTransactionContext
from tealer.analyses.dataflow.transaction_context.utils.key_helpers import (
    get_gtxn_at_index_key,
    is_value_matches_key,
    get_absolute_index_key,
    get_relative_index_key,
)
from tealer.teal.instructions.instructions import (
    Eq,
    Neq,
    Greater,
    GreaterE,
    Less,
    LessE,
)
from tealer.utils.analyses import is_int_push_ins
from tealer.utils.algorand_constants import MAX_GROUP_SIZE, MAX_UINT64, MAX_TRANSACTION_COST
from tealer.analyses.utils.stack_ast_builder import KnownStackValue, UnknownStackValue

if TYPE_CHECKING:
    from tealer.teal.instructions.instructions import Instruction
    from tealer.teal.basic_blocks import BasicBlock

# It's not always possible to know the integer value used to compare Fee.
# All the unknown values are represented using FeeValue(is_unknown=True).
# Tealer does not report paths if Fee is less than MAX_TRANSACTION_COST.
# UnknownBoundedInt is assumed to be a value between 0 and MAX_TRANSACTION_COST
# Tealer does not report paths if Fee is checked against some unknown value.

FEE_KEY = "Fee"


# TODO: Change is_unknown to unknown/known/unbounded enum.
@dataclass
class FeeValue:
    is_unknown: bool = False  # max(unknown_value) == MAX_TRANSACTION_COST
    value: int = MAX_UINT64  # use value if and only if is_unknown is False.


# Note: if Fee field analysis is only used by the detector, it might be enough to represent unknown value with int(MAX_TRANSACTION_COST)
# However, if Fee field analysis is used by other components or tools(Fuzzer?), it is important to differentiate unknown and known values.


class FeeField(DataflowTransactionContext):

    BASE_KEYS: List[str] = [FEE_KEY]
    KEYS_WITH_GTXN: List[str] = [FEE_KEY]

    def _universal_set(self, key: str) -> FeeValue:
        # MAX possible value for the Fee field
        return FeeValue()

    def _null_set(self, key: str) -> FeeValue:
        return FeeValue(value=0)

    # def _union(self, key: str, a: Union[int, str], b: Union[int, str]) -> Union[int, str]:
    def _union(self, key: str, a: FeeValue, b: FeeValue) -> FeeValue:
        if a.is_unknown and b.is_unknown:
            # Both are unknown values between 0 < x <= MAX_TRANSACTION_COST
            return a
        if a.is_unknown:
            # a is a unknown bounded value and b is known value
            # a: 0 < x < MAX_TRANSACTION_COST
            if b.value > MAX_TRANSACTION_COST:
                return b
            return a  # a is less than MAX_TRANSACTION_COST
        if b.is_unknown:
            # b is a unknown bounded value and a is known value
            # b: 0 < x < MAX_TRANSACTION_COST
            if a.value > MAX_TRANSACTION_COST:
                return a  # b is less than MAX_TRANSACTION_COST
            return b
        # both are known values
        return a if a.value > b.value else b

    def _intersection(self, key: str, a: FeeValue, b: FeeValue) -> FeeValue:
        if a.is_unknown and b.is_unknown:
            # Both are unknown values between 0 < x < MAX_TRANSACTION_COST
            return a
        if a.is_unknown:
            # a is a unknown bounded value and b is int
            # a: 0 < x < MAX_TRANSACTION_COST
            if b.value > MAX_TRANSACTION_COST:
                return a  # a is less than MAX_TRANSACTION_COST
            return b
        if b.is_unknown:
            # b is a unknown bounded value and a is int
            # b: 0 < x < MAX_TRANSACTION_COST
            if a.value > MAX_TRANSACTION_COST:
                return b  # b is less than MAX_TRANSACTION_COST
            return a
        # both are known values
        return a if a.value < b.value else b

    @staticmethod
    def _get_asserted_max_value(
        comparison_ins: "Instruction", compared_value: FeeValue
    ) -> Tuple[FeeValue, FeeValue]:
        """Return maximum possible value that will make the comparison True and maximum
        possible value that will make the comparison False. Both values are upper bounded
        by arg "max_possible_value".

        Args:
            comparison_ins: Comparison operator.
            compared_value: fee value being compared with.

        Returns:
            Max possible value that will make the comparison instruction return True and
                Max possible value that will make the comparison False.
        """
        # U = max_possible_value  # universal set
        if isinstance(comparison_ins, Eq):
            # x == i => i, U
            return compared_value, FeeValue()
        if isinstance(comparison_ins, Neq):
            # x != i => U, i
            return FeeValue(), compared_value
        if isinstance(comparison_ins, Less):
            # x < i => (i - 1), U
            if compared_value.is_unknown:
                return compared_value, FeeValue()
            return FeeValue(value=max(0, compared_value.value - 1)), FeeValue()
        if isinstance(comparison_ins, LessE):
            # x <= i => i, U
            return compared_value, FeeValue()
        if isinstance(comparison_ins, Greater):
            # x > i => U, i
            return FeeValue(), compared_value
        if isinstance(comparison_ins, GreaterE):
            # x >= i => U, (i - 1)
            if compared_value.is_unknown:
                return FeeValue(), compared_value
            return FeeValue(), FeeValue(value=max(0, compared_value.value - 1))
        return FeeValue(), FeeValue()

    def _get_asserted_fee(  # pylint: disable=too-many-branches
        self, key: str, ins_stack_value: KnownStackValue
    ) -> Tuple[FeeValue, FeeValue]:

        if isinstance(ins_stack_value.instruction, (Eq, Neq, Less, LessE, Greater, GreaterE)):
            arg1 = ins_stack_value.args[0]
            arg2 = ins_stack_value.args[1]
            compared_value: Optional[FeeValue] = None

            if isinstance(arg1, UnknownStackValue) and isinstance(arg2, UnknownStackValue):
                # Both the args are unknown
                # return U, U
                return FeeValue(), FeeValue()

            if isinstance(arg1, UnknownStackValue):
                if not isinstance(arg2, UnknownStackValue) and not is_value_matches_key(key, arg2):
                    # arg1 is unknown and arg2 is not related to "key"
                    return FeeValue(), FeeValue()
                # arg2 is related to key and arg1 is some unknown value
                compared_value = FeeValue(is_unknown=True)
            elif isinstance(arg2, UnknownStackValue):
                if not isinstance(arg1, UnknownStackValue) and not is_value_matches_key(key, arg1):
                    # arg2 is unknown and arg1 is not related to "key"
                    return FeeValue(), FeeValue()
                # arg1 is related to "key" and arg2 is some unknown int value
                compared_value = FeeValue(is_unknown=True)
            elif is_value_matches_key(key, arg1):
                is_int, value = is_int_push_ins(arg2.instruction)
                if is_int and isinstance(value, int):
                    compared_value = FeeValue(value=value)
                else:
                    compared_value = FeeValue(is_unknown=True)

            elif is_value_matches_key(key, arg2):
                is_int, value = is_int_push_ins(arg1.instruction)
                if is_int and isinstance(value, int):
                    compared_value = FeeValue(value=value)
                else:
                    compared_value = FeeValue(is_unknown=True)

            if compared_value is None:
                # compared_value is not int.
                return FeeValue(), FeeValue()

            ins = ins_stack_value.instruction
            return self._get_asserted_max_value(ins, compared_value)
        return FeeValue(), FeeValue()

    def _get_asserted_single(
        self, key: str, ins_stack_value: KnownStackValue
    ) -> Tuple[FeeValue, FeeValue]:
        res = self._get_asserted_fee(key, ins_stack_value)
        return res

    def _store_results(self) -> None:
        for block in self._function.blocks:
            max_fee = self._block_contexts[FEE_KEY][block]
            assert isinstance(max_fee, FeeValue)
            if max_fee.is_unknown:
                self._function.transaction_context(block).max_fee_unknown = True
            else:
                self._function.transaction_context(block).max_fee = max_fee.value
            for idx in range(MAX_GROUP_SIZE):
                max_fee = self._block_contexts[get_gtxn_at_index_key(idx, FEE_KEY)][block]
                assert isinstance(max_fee, FeeValue)
                if max_fee.is_unknown:
                    self._function.transaction_context(block).gtxn_context(
                        idx
                    ).max_fee_unknown = True
                else:
                    self._function.transaction_context(block).gtxn_context(
                        idx
                    ).max_fee = max_fee.value

                abs_max_fee = self._block_contexts[get_absolute_index_key(idx, FEE_KEY)][block]
                assert isinstance(abs_max_fee, FeeValue)
                if abs_max_fee.is_unknown:
                    self._function.transaction_context(block).absolute_context(
                        idx
                    ).max_fee_unknown = True
                else:
                    self._function.transaction_context(block).absolute_context(
                        idx
                    ).max_fee = abs_max_fee.value

            for offset in range(-(MAX_GROUP_SIZE - 1), MAX_GROUP_SIZE):
                if offset == 0:
                    continue
                rel_max_fee = self._block_contexts[get_relative_index_key(offset, FEE_KEY)][block]
                assert isinstance(rel_max_fee, FeeValue)
                if rel_max_fee.is_unknown:
                    self._function.transaction_context(block).relative_context(
                        offset
                    ).max_fee_unknown = True
                else:
                    self._function.transaction_context(block).relative_context(
                        offset
                    ).max_fee = rel_max_fee.value



"""Collects and exports detectors defined in tealer.

Exported detectors are:

* ``isDeletable``: IsDeletable detects paths missing DeleteApplication check.
* ``isUpdatable``: IsUpdatable detects paths missing UpdateApplication check.
* ``groupSize``: MissingGroupSize detects paths missing GroupSize check.
* ``rekeyTo``: MissingRekeyTo detects paths missing RekeyTo check.
* ``canCloseAccount``: CanCloseAccount detects paths missing CloseRemainderTo check.
* ``canCloseAsset``: CanCloseAsset detects paths missing AssetCloseTo check.
* ``feeCheck``: MissingFeeCheck detects paths missing Fee check.
* ``anyoneCanUpdate``: Detect paths missing validations on sender field AND allows to delete the application.
* ``anyoneCanDelete``: Detect paths missing validations on sender field AND allows to update the application.
"""

# pylint: disable=unused-import
from tealer.detectors.is_deletable import IsDeletable
from tealer.detectors.is_updatable import IsUpdatable
from tealer.detectors.groupsize import MissingGroupSize
from tealer.detectors.rekeyto import MissingRekeyTo
from tealer.detectors.can_close_account import CanCloseAccount
from tealer.detectors.can_close_asset import CanCloseAsset
from tealer.detectors.fee_check import MissingFeeCheck
from tealer.detectors.anyone_can_update import AnyoneCanUpdate
from tealer.detectors.anyone_can_delete import AnyoneCanDelete
from tealer.detectors.optimizations.constant_gtxn import ConstantGtxn
from tealer.detectors.optimizations.sender_access import SenderAccess
from tealer.detectors.optimizations.self_access import SelfAccess



from typing import Tuple, List
import pytest

from tealer.analyses.utils.stack_ast_builder import (
    construct_stack_ast,
    UnknownStackValue,
    KnownStackValue,
    StackValue,
)
from tealer.teal.parse_teal import parse_teal
from tealer.teal.instructions.instructions import (
    Pragma,
    Txn,
    Int,
    Neq,
    Gtxn,
    Addr,
    Eq,
    B,
    Global,
    Label,
    PushInt,
    LessE,
    Mul,
    PushBytes,
    Assert,
    And,
    Or,
    BZ,
    BNZ,
    Return,
)
from tealer.teal.instructions.transaction_field import (
    GroupIndex,
    CloseRemainderTo,
    TypeEnum,
    RekeyTo,
    LastValid,
    Fee,
    Lease,
)
from tealer.teal.global_field import GroupSize, ZeroAddress, MinTxnFee


T1 = """
#pragma version 6
txn GroupIndex
int 0
!=
bnz index_not_0
gtxn 0 CloseRemainderTo  // index must be 0
addr AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAEVAL4QAJS7JHB4
==
assert
b process_txn
index_not_0:
global GroupSize    // index is not zero and groupsize is **3**. So, txn index can be 1 or 2
int 3
==
assert
gtxn 1 CloseRemainderTo     // only checks gtxn 1. so, is vulnerable if txn index is 2
global ZeroAddress
==
assert
process_txn:
int 1
return
"""

T1_VALUES = [
    KnownStackValue(ins=Pragma(6), args=[]),
    KnownStackValue(ins=Txn(GroupIndex()), args=[]),
    KnownStackValue(ins=Int(0), args=[]),
    KnownStackValue(
        ins=Neq(),
        args=[
            KnownStackValue(ins=Txn(GroupIndex()), args=[]),
            KnownStackValue(ins=Int(0), args=[]),
        ],
    ),
    KnownStackValue(
        ins=BNZ("index_not_0"),
        args=[
            KnownStackValue(
                ins=Neq(),
                args=[
                    KnownStackValue(ins=Txn(GroupIndex()), args=[]),
                    KnownStackValue(ins=Int(0), args=[]),
                ],
            )
        ],
    ),
    KnownStackValue(ins=Gtxn(0, CloseRemainderTo()), args=[]),
    KnownStackValue(
        ins=Addr("AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAEVAL4QAJS7JHB4"), args=[]
    ),
    KnownStackValue(
        ins=Eq(),
        args=[
            KnownStackValue(ins=Gtxn(0, CloseRemainderTo()), args=[]),
            KnownStackValue(
                ins=Addr("AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAEVAL4QAJS7JHB4"), args=[]
            ),
        ],
    ),
    KnownStackValue(
        ins=Assert(),
        args=[
            KnownStackValue(
                ins=Eq(),
                args=[
                    KnownStackValue(ins=Gtxn(0, CloseRemainderTo()), args=[]),
                    KnownStackValue(
                        ins=Addr("AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAEVAL4QAJS7JHB4"),
                        args=[],
                    ),
                ],
            )
        ],
    ),
    KnownStackValue(ins=B("process_txn"), args=[]),
    KnownStackValue(ins=Label("index_not_0:"), args=[]),
    KnownStackValue(ins=Global(GroupSize()), args=[]),
    KnownStackValue(ins=Int(3), args=[]),
    KnownStackValue(
        ins=Eq(),
        args=[
            KnownStackValue(ins=Global(GroupSize()), args=[]),
            KnownStackValue(ins=Int(3), args=[]),
        ],
    ),
    KnownStackValue(
        ins=Assert(),
        args=[
            KnownStackValue(
                ins=Eq(),
                args=[
                    KnownStackValue(ins=Global(GroupSize()), args=[]),
                    KnownStackValue(ins=Int(3), args=[]),
                ],
            )
        ],
    ),
    KnownStackValue(ins=Gtxn(1, CloseRemainderTo()), args=[]),
    KnownStackValue(ins=Global(ZeroAddress()), args=[]),
    KnownStackValue(
        ins=Eq(),
        args=[
            KnownStackValue(ins=Gtxn(1, CloseRemainderTo()), args=[]),
            KnownStackValue(ins=Global(ZeroAddress()), args=[]),
        ],
    ),
    KnownStackValue(
        ins=Assert(),
        args=[
            KnownStackValue(
                ins=Eq(),
                args=[
                    KnownStackValue(ins=Gtxn(1, CloseRemainderTo()), args=[]),
                    KnownStackValue(ins=Global(ZeroAddress()), args=[]),
                ],
            )
        ],
    ),
    KnownStackValue(ins=Label("process_txn:"), args=[]),
    KnownStackValue(ins=Int(1), args=[]),
    KnownStackValue(ins=Return(), args=[KnownStackValue(ins=Int(1), args=[])]),
]

T2 = """
#pragma version 4
txn TypeEnum
int 2
==
txn RekeyTo
global ZeroAddress
==
&&
txn LastValid
pushint 16600000
==
&&
txn Fee
global MinTxnFee
<=
&&
txn Lease
pushbytes 0x0000000000000000000000000000000000000000000000000000000000000001
==
&&
bz main_l2
int 1
return
main_l2:
txn RekeyTo
global ZeroAddress
==
||
txn CloseRemainderTo
global ZeroAddress
==
||
txn Fee
int 2
global MinTxnFee
*
<=
||
assert
int 1
return
"""

T2_VALUES = [
    KnownStackValue(ins=Pragma(4), args=[]),
    KnownStackValue(ins=Txn(TypeEnum()), args=[]),
    KnownStackValue(ins=Int(2), args=[]),
    KnownStackValue(
        ins=Eq(),
        args=[KnownStackValue(ins=Txn(TypeEnum()), args=[]), KnownStackValue(ins=Int(2), args=[])],
    ),
    KnownStackValue(ins=Txn(RekeyTo()), args=[]),
    KnownStackValue(ins=Global(ZeroAddress()), args=[]),
    KnownStackValue(
        ins=Eq(),
        args=[
            KnownStackValue(ins=Txn(RekeyTo()), args=[]),
            KnownStackValue(ins=Global(ZeroAddress()), args=[]),
        ],
    ),
    KnownStackValue(
        ins=And(),
        args=[
            KnownStackValue(
                ins=Eq(),
                args=[
                    KnownStackValue(ins=Txn(TypeEnum()), args=[]),
                    KnownStackValue(ins=Int(2), args=[]),
                ],
            ),
            KnownStackValue(
                ins=Eq(),
                args=[
                    KnownStackValue(ins=Txn(RekeyTo()), args=[]),
                    KnownStackValue(ins=Global(ZeroAddress()), args=[]),
                ],
            ),
        ],
    ),
    KnownStackValue(ins=Txn(LastValid()), args=[]),
    KnownStackValue(ins=PushInt(16600000), args=[]),
    KnownStackValue(
        ins=Eq(),
        args=[
            KnownStackValue(ins=Txn(LastValid()), args=[]),
            KnownStackValue(ins=PushInt(16600000), args=[]),
        ],
    ),
    KnownStackValue(
        ins=And(),
        args=[
            KnownStackValue(
                ins=And(),
                args=[
                    KnownStackValue(
                        ins=Eq(),
                        args=[
                            KnownStackValue(ins=Txn(TypeEnum()), args=[]),
                            KnownStackValue(ins=Int(2), args=[]),
                        ],
                    ),
                    KnownStackValue(
                        ins=Eq(),
                        args=[
                            KnownStackValue(ins=Txn(RekeyTo()), args=[]),
                            KnownStackValue(ins=Global(ZeroAddress()), args=[]),
                        ],
                    ),
                ],
            ),
            KnownStackValue(
                ins=Eq(),
                args=[
                    KnownStackValue(ins=Txn(LastValid()), args=[]),
                    KnownStackValue(ins=PushInt(16600000), args=[]),
                ],
            ),
        ],
    ),
    KnownStackValue(ins=Txn(Fee()), args=[]),
    KnownStackValue(ins=Global(MinTxnFee()), args=[]),
    KnownStackValue(
        ins=LessE(),
        args=[
            KnownStackValue(ins=Txn(Fee()), args=[]),
            KnownStackValue(ins=Global(MinTxnFee()), args=[]),
        ],
    ),
    KnownStackValue(
        ins=And(),
        args=[
            KnownStackValue(
                ins=And(),
                args=[
                    KnownStackValue(
                        ins=And(),
                        args=[
                            KnownStackValue(
                                ins=Eq(),
                                args=[
                                    KnownStackValue(ins=Txn(TypeEnum()), args=[]),
                                    KnownStackValue(ins=Int(2), args=[]),
                                ],
                            ),
                            KnownStackValue(
                                ins=Eq(),
                                args=[
                                    KnownStackValue(ins=Txn(RekeyTo()), args=[]),
                                    KnownStackValue(ins=Global(ZeroAddress()), args=[]),
                                ],
                            ),
                        ],
                    ),
                    KnownStackValue(
                        ins=Eq(),
                        args=[
                            KnownStackValue(ins=Txn(LastValid()), args=[]),
                            KnownStackValue(ins=PushInt(16600000), args=[]),
                        ],
                    ),
                ],
            ),
            KnownStackValue(
                ins=LessE(),
                args=[
                    KnownStackValue(ins=Txn(Fee()), args=[]),
                    KnownStackValue(ins=Global(MinTxnFee()), args=[]),
                ],
            ),
        ],
    ),
    KnownStackValue(ins=Txn(Lease()), args=[]),
    KnownStackValue(
        ins=PushBytes("0x0000000000000000000000000000000000000000000000000000000000000001"), args=[]
    ),
    KnownStackValue(
        ins=Eq(),
        args=[
            KnownStackValue(ins=Txn(Lease()), args=[]),
            KnownStackValue(
                ins=PushBytes("0x0000000000000000000000000000000000000000000000000000000000000001"),
                args=[],
            ),
        ],
    ),
    KnownStackValue(
        ins=And(),
        args=[
            KnownStackValue(
                ins=And(),
                args=[
                    KnownStackValue(
                        ins=And(),
                        args=[
                            KnownStackValue(
                                ins=And(),
                                args=[
                                    KnownStackValue(
                                        ins=Eq(),
                                        args=[
                                            KnownStackValue(ins=Txn(TypeEnum()), args=[]),
                                            KnownStackValue(ins=Int(2), args=[]),
                                        ],
                                    ),
                                    KnownStackValue(
                                        ins=Eq(),
                                        args=[
                                            KnownStackValue(ins=Txn(RekeyTo()), args=[]),
                                            KnownStackValue(ins=Global(ZeroAddress()), args=[]),
                                        ],
                                    ),
                                ],
                            ),
                            KnownStackValue(
                                ins=Eq(),
                                args=[
                                    KnownStackValue(ins=Txn(LastValid()), args=[]),
                                    KnownStackValue(ins=PushInt(16600000), args=[]),
                                ],
                            ),
                        ],
                    ),
                    KnownStackValue(
                        ins=LessE(),
                        args=[
                            KnownStackValue(ins=Txn(Fee()), args=[]),
                            KnownStackValue(ins=Global(MinTxnFee()), args=[]),
                        ],
                    ),
                ],
            ),
            KnownStackValue(
                ins=Eq(),
                args=[
                    KnownStackValue(ins=Txn(Lease()), args=[]),
                    KnownStackValue(
                        ins=PushBytes(
                            "0x0000000000000000000000000000000000000000000000000000000000000001"
                        ),
                        args=[],
                    ),
                ],
            ),
        ],
    ),
    KnownStackValue(
        ins=BZ("main_l2"),
        args=[
            KnownStackValue(
                ins=And(),
                args=[
                    KnownStackValue(
                        ins=And(),
                        args=[
                            KnownStackValue(
                                ins=And(),
                                args=[
                                    KnownStackValue(
                                        ins=And(),
                                        args=[
                                            KnownStackValue(
                                                ins=Eq(),
                                                args=[
                                                    KnownStackValue(ins=Txn(TypeEnum()), args=[]),
                                                    KnownStackValue(ins=Int(2), args=[]),
                                                ],
                                            ),
                                            KnownStackValue(
                                                ins=Eq(),
                                                args=[
                                                    KnownStackValue(ins=Txn(RekeyTo()), args=[]),
                                                    KnownStackValue(
                                                        ins=Global(ZeroAddress()), args=[]
                                                    ),
                                                ],
                                            ),
                                        ],
                                    ),
                                    KnownStackValue(
                                        ins=Eq(),
                                        args=[
                                            KnownStackValue(ins=Txn(LastValid()), args=[]),
                                            KnownStackValue(ins=PushInt(16600000), args=[]),
                                        ],
                                    ),
                                ],
                            ),
                            KnownStackValue(
                                ins=LessE(),
                                args=[
                                    KnownStackValue(ins=Txn(Fee()), args=[]),
                                    KnownStackValue(ins=Global(MinTxnFee()), args=[]),
                                ],
                            ),
                        ],
                    ),
                    KnownStackValue(
                        ins=Eq(),
                        args=[
                            KnownStackValue(ins=Txn(Lease()), args=[]),
                            KnownStackValue(
                                ins=PushBytes(
                                    "0x0000000000000000000000000000000000000000000000000000000000000001"
                                ),
                                args=[],
                            ),
                        ],
                    ),
                ],
            ),
        ],
    ),
    KnownStackValue(ins=Int(1), args=[]),
    KnownStackValue(ins=Return(), args=[KnownStackValue(ins=Int(1), args=[])]),
    KnownStackValue(ins=Label("main_l2:"), args=[]),
    KnownStackValue(ins=Txn(RekeyTo()), args=[]),
    KnownStackValue(ins=Global(ZeroAddress()), args=[]),
    KnownStackValue(
        ins=Eq(),
        args=[
            KnownStackValue(ins=Txn(RekeyTo()), args=[]),
            KnownStackValue(ins=Global(ZeroAddress()), args=[]),
        ],
    ),
    KnownStackValue(
        ins=Or(),
        args=[
            UnknownStackValue(),
            KnownStackValue(
                ins=Eq(),
                args=[
                    KnownStackValue(ins=Txn(RekeyTo()), args=[]),
                    KnownStackValue(ins=Global(ZeroAddress()), args=[]),
                ],
            ),
        ],
    ),
    KnownStackValue(ins=Txn(CloseRemainderTo()), args=[]),
    KnownStackValue(ins=Global(ZeroAddress()), args=[]),
    KnownStackValue(
        ins=Eq(),
        args=[
            KnownStackValue(ins=Txn(CloseRemainderTo()), args=[]),
            KnownStackValue(ins=Global(ZeroAddress()), args=[]),
        ],
    ),
    KnownStackValue(
        ins=Or(),
        args=[
            KnownStackValue(
                ins=Or(),
                args=[
                    UnknownStackValue(),
                    KnownStackValue(
                        ins=Eq(),
                        args=[
                            KnownStackValue(ins=Txn(RekeyTo()), args=[]),
                            KnownStackValue(ins=Global(ZeroAddress()), args=[]),
                        ],
                    ),
                ],
            ),
            KnownStackValue(
                ins=Eq(),
                args=[
                    KnownStackValue(ins=Txn(CloseRemainderTo()), args=[]),
                    KnownStackValue(ins=Global(ZeroAddress()), args=[]),
                ],
            ),
        ],
    ),
    KnownStackValue(ins=Txn(Fee()), args=[]),
    KnownStackValue(ins=Int(2), args=[]),
    KnownStackValue(ins=Global(MinTxnFee()), args=[]),
    KnownStackValue(
        ins=Mul(),
        args=[
            KnownStackValue(ins=Int(2), args=[]),
            KnownStackValue(ins=Global(MinTxnFee()), args=[]),
        ],
    ),
    KnownStackValue(
        ins=LessE(),
        args=[
            KnownStackValue(ins=Txn(Fee()), args=[]),
            KnownStackValue(
                ins=Mul(),
                args=[
                    KnownStackValue(ins=Int(2), args=[]),
                    KnownStackValue(ins=Global(MinTxnFee()), args=[]),
                ],
            ),
        ],
    ),
    KnownStackValue(
        ins=Or(),
        args=[
            KnownStackValue(
                ins=Or(),
                args=[
                    KnownStackValue(
                        ins=Or(),
                        args=[
                            UnknownStackValue(),
                            KnownStackValue(
                                ins=Eq(),
                                args=[
                                    KnownStackValue(ins=Txn(RekeyTo()), args=[]),
                                    KnownStackValue(ins=Global(ZeroAddress()), args=[]),
                                ],
                            ),
                        ],
                    ),
                    KnownStackValue(
                        ins=Eq(),
                        args=[
                            KnownStackValue(ins=Txn(CloseRemainderTo()), args=[]),
                            KnownStackValue(ins=Global(ZeroAddress()), args=[]),
                        ],
                    ),
                ],
            ),
            KnownStackValue(
                ins=LessE(),
                args=[
                    KnownStackValue(ins=Txn(Fee()), args=[]),
                    KnownStackValue(
                        ins=Mul(),
                        args=[
                            KnownStackValue(ins=Int(2), args=[]),
                            KnownStackValue(ins=Global(MinTxnFee()), args=[]),
                        ],
                    ),
                ],
            ),
        ],
    ),
    KnownStackValue(
        ins=Assert(),
        args=[
            KnownStackValue(
                ins=Or(),
                args=[
                    KnownStackValue(
                        ins=Or(),
                        args=[
                            KnownStackValue(
                                ins=Or(),
                                args=[
                                    UnknownStackValue(),
                                    KnownStackValue(
                                        ins=Eq(),
                                        args=[
                                            KnownStackValue(ins=Txn(RekeyTo()), args=[]),
                                            KnownStackValue(ins=Global(ZeroAddress()), args=[]),
                                        ],
                                    ),
                                ],
                            ),
                            KnownStackValue(
                                ins=Eq(),
                                args=[
                                    KnownStackValue(ins=Txn(CloseRemainderTo()), args=[]),
                                    KnownStackValue(ins=Global(ZeroAddress()), args=[]),
                                ],
                            ),
                        ],
                    ),
                    KnownStackValue(
                        ins=LessE(),
                        args=[
                            KnownStackValue(ins=Txn(Fee()), args=[]),
                            KnownStackValue(
                                ins=Mul(),
                                args=[
                                    KnownStackValue(ins=Int(2), args=[]),
                                    KnownStackValue(ins=Global(MinTxnFee()), args=[]),
                                ],
                            ),
                        ],
                    ),
                ],
            ),
        ],
    ),
    KnownStackValue(ins=Int(1), args=[]),
    KnownStackValue(
        ins=Return(),
        args=[
            KnownStackValue(ins=Int(1), args=[]),
        ],
    ),
]


ALL_NEW_TESTS = [
    (T1, T1_VALUES),
    (T2, T2_VALUES),
]


def cmp_stack_values(ex_v: StackValue, t_v: StackValue) -> bool:
    if isinstance(ex_v, UnknownStackValue):
        print("t_v", t_v)
        return isinstance(t_v, UnknownStackValue)
    if isinstance(t_v, UnknownStackValue):
        print("t_v", t_v)
        return isinstance(ex_v, UnknownStackValue)
    if ex_v.instruction.__class__ != t_v.instruction.__class__:
        return False
    if len(ex_v.args) != len(t_v.args):
        return False
    for ex_arg, t_arg in zip(ex_v.args, t_v.args):
        if not cmp_stack_values(ex_arg, t_arg):
            return False
    return True


@pytest.mark.parametrize("test", ALL_NEW_TESTS)  # type: ignore
def test_just_detectors(test: Tuple[str, List[KnownStackValue]]) -> None:
    code, expected_stack_values = test
    teal = parse_teal(code.strip())
    test_values: List[KnownStackValue] = []
    for bi in teal.bbs:
        values = construct_stack_ast(bi)
        for ins in bi.instructions:
            test_values.append(values[ins])
    assert len(expected_stack_values) == len(test_values)
    for ex_v, t_v in zip(expected_stack_values, test_values):
        print(str(t_v.instruction), t_v.instruction.line)
        assert cmp_stack_values(ex_v, t_v)


if __name__ == "__main__":
    teal_obj = parse_teal(T2)
    for bb in teal_obj.bbs:
        print(bb)
        result = construct_stack_ast(bb)
        for i in bb.instructions:
            print(i, result[i])
        print("----------" * 10)



"""Module for tealer detectors.

Detectors are analyzers which go through the parsed code and look for
certain patterns in the program which are of interest to anyone invested
in the security of the contract.

This module contains Abstract base class for printers and implementation of
few detectors that detect common issues in algorand smart contracts.

Modules:
    abstract_detector: Defines abstract base class for detectors.

    all_detectors: Collects and exports detectors defined in tealer.
"""



"""Parser for fields of asset_holding_get.

Each ``asset_holding_get`` field is represented as a class. Parsing
the field is creating the class instance representing the field given
it's string representation.

Fields of ``asset_holding_get`` doesn't have immediate arguments
and have very simple structure. All fields map one-to-one with
their string representation. As a result, asset holding fields are
parsed by first constructing a map(dict) from their string
representation to corresponding class representing the field and
using that to find the correct class for the given field.

Attributes:
    ASSET_HOLDING_FIELD_TXT_TO_OBJECT: Map(dict) from string representation
        of asset holding field to the corresponding class.
"""

from tealer.teal.instructions.asset_holding_field import (
    AssetHoldingField,
    AssetBalance,
    AssetFrozen,
)

ASSET_HOLDING_FIELD_TXT_TO_OBJECT = {"AssetBalance": AssetBalance, "AssetFrozen": AssetFrozen}


def parse_asset_holding_field(field: str) -> AssetHoldingField:
    """Parse fields of ``asset_holding_get``.

    Args:
        field: string representation of the field.

    Returns:
        object of class corresponding to the given asset_holding_get field.
    """

    return ASSET_HOLDING_FIELD_TXT_TO_OBJECT[field]()



"""Module for tealer printers.

Printers are information summarizers. Printers make finding and evaluating
interesting information related to the contract easier. There's always abundant
information that can be extracted from a smart contract, manually going
through a smart contract for a every tiny bit of information is hard, boringgg
and error prone. Printers make this process of finding information easier.

This module contains Abstract base class for printers and implementation of
few printers that can used to retrieve interesting information.

Modules:
    abstract_printer: Defines abstract base class for printers in tealer.

    all_printers: Collects and exports printers defined in tealer.

    call_graph: Printer for exporting call-graph of the contract.

    function_cfg: Printer for exporting contract subroutines in dot format.

    human_summary: Printer to print summary of the contract.
"""



# pylint: skip-file
# pylint: disable=undefined-variable
# type: ignore[name-defined]
from pyteal import *  # pylint: disable=wildcard-import, unused-wildcard-import


@Subroutine(TealType.uint64)
def add_upto_n(n: ScratchVar) -> Expr:
    i = ScratchVar()
    result = ScratchVar()
    return Seq(
        [
            result.store(Int(0)),
            For(i.store(Int(0)), i.load() < n.load(), i.store(i.load() + Int(1))).Do(
                result.store(result.load() + i.load())
            ),
            Return(result.load()),
        ]
    )


def on_creation() -> Expr:
    x = ScratchVar()
    return Seq([x.store(Int(10)), Pop(add_upto_n(x)), Return(Int(1))])


def approval_program() -> Expr:
    return Cond(
        [Txn.application_id() == Int(0), on_creation()],
        [Txn.on_completion() == OnComplete.DeleteApplication, Reject()],
        [Txn.on_completion() == OnComplete.UpdateApplication, Reject()],
        [Txn.on_completion() == OnComplete.CloseOut, Approve()],
        [Txn.on_completion() == OnComplete.OptIn, Approve()],
    )


def clear_program() -> Expr:
    return Return(Int(1))


normal_application_approval_program = compileTeal(
    approval_program(),
    mode=Mode.Application,
    version=7,
    optimize=OptimizeOptions(scratch_slots=False),
    assembleConstants=True,
)

_clear_program = compileTeal(
    clear_program(), mode=Mode.Application, version=7, optimize=OptimizeOptions(scratch_slots=False)
)

if __name__ == "__main__":
    print(normal_application_approval_program)



from typing import List, Tuple
import pytest

from tealer.utils.command_line.common import init_tealer_from_single_contract
from tests.utils import order_basic_blocks


MULTIPLE_RETSUB = """
#pragma version 5
b main
is_even:
    int 2
    %
    bz return_1
    int 0
    txn GroupIndex
    int 2
    ==
    assert
    retsub
return_1:
    txn GroupIndex
    int 3
    <
    assert
    int 1
    retsub
main:
    txn GroupIndex
    int 1
    !=
    assert
    int 4
    callsub is_even
    return
"""

MULTIPLE_RETSUB_GROUP_INDICES = [[0, 2], [0, 2], [2], [0, 2], [0, 2], [0, 2]]

SUBROUTINE_BACK_JUMP = """
#pragma version 5
b main
getmod:
    %
    retsub
is_odd:
    txn GroupIndex
    int 4
    <
    assert
    txn GroupIndex
    int 2
    !=
    assert
    int 2
    b getmod
main:
    int 5
    callsub is_odd
    return
"""

SUBROUTINE_BACK_JUMP_GROUP_INDICES = [
    [0, 1, 3],
    [0, 1, 3],
    [0, 1, 3],
    [0, 1, 3],
    [0, 1, 3],
    [0, 1, 3],
]

BRANCHING = """
#pragma version 4
txn GroupIndex
int 2
>=
assert
txn GroupIndex
int 4
>
bz fin
txn GroupIndex
int 1
==
bnz check_second_arg
int 0
return
check_second_arg:
txn ApplicationArgs 1
btoi
int 100
>
bnz fin
int 0
return
fin:
int 1
return
"""

BRANCHING_GROUP_INDICES = [[2, 3, 4], [], [], [], [], [2, 3, 4]]

LOOPS = """
#pragma version 5
txn GroupIndex
int 4
!=
assert
int 0
loop:
    dup
    txn GroupIndex
    int 3
    >=
    bz end
    int 1
    +
    txn GroupIndex
    int 3
    <
    assert
    b loop
end:
    int 2
    txn GroupIndex
    ==
    assert
    int 1
    return
"""

LOOPS_GROUP_INDICES = [[2], [2], [], [2]]

LOOPS_GROUP_SIZES = """
#pragma version 5
txn GroupIndex
int 4
!=
assert
global GroupSize
int 6
<=
int 0
loop:
    dup
    txn GroupIndex
    int 3
    >
    bz end
    int 1
    +
    txn GroupIndex
    int 6
    <
    assert
    b loop
end:
    int 2
    txn GroupIndex
    >
    assert
    int 5
    global GroupSize
    <=
    assert
    int 1
    return
"""

LOOPS_GROUP_SIZES_GROUP_INDICES = [[3], [3], [], [3]]

ALL_TESTS = [
    (MULTIPLE_RETSUB, MULTIPLE_RETSUB_GROUP_INDICES),
    (SUBROUTINE_BACK_JUMP, SUBROUTINE_BACK_JUMP_GROUP_INDICES),
    (BRANCHING, BRANCHING_GROUP_INDICES),
    (LOOPS, LOOPS_GROUP_INDICES),
    (LOOPS_GROUP_SIZES, LOOPS_GROUP_SIZES_GROUP_INDICES),
]


@pytest.mark.parametrize("test", ALL_TESTS)  # type: ignore
def test_group_indices(test: Tuple[str, List[List[int]]]) -> None:
    code, group_indices = test
    tealer = init_tealer_from_single_contract(code.strip(), "test")
    function = tealer.contracts["test"].functions["test"]

    bbs = order_basic_blocks(function.blocks)
    for b, indices in zip(bbs, group_indices):
        assert function.transaction_context(b).group_indices == indices



from typing import TYPE_CHECKING, List, Set, Tuple, Callable

from tealer.analyses.dataflow.transaction_context.generic import DataflowTransactionContext
from tealer.analyses.dataflow.transaction_context.utils.key_helpers import (
    get_gtxn_at_index_key,
    is_value_matches_key,
    get_absolute_index_key,
    get_relative_index_key,
)
from tealer.teal.instructions.instructions import (
    Eq,
    Neq,
    Addr,
    Global,
)
from tealer.teal.global_field import ZeroAddress, CreatorAddress
from tealer.utils.algorand_constants import ZERO_ADDRESS
from tealer.analyses.utils.stack_ast_builder import KnownStackValue, UnknownStackValue
from tealer.utils.algorand_constants import MAX_GROUP_SIZE

if TYPE_CHECKING:
    from tealer.teal.instructions.instructions import Instruction
    from tealer.teal.basic_blocks import BasicBlock
    from tealer.teal.context.block_transaction_context import (
        AddrFieldValue,
        BlockTransactionContext,
    )


# TODO: Change representation of address values to something similar to FeeValue
# - Store possible addresses and not possible address. Neq(Txn Sender, addr x). store Sender cannot be x
# - ...

# Add str of transaction field to find constraints on that field
# use field string as in TX_FIELD_TXT_TO_OBJECT
REKEY_TO_KEY = "RekeyTo"
CLOSE_REMAINDER_TO_KEY = "CloseRemainderTo"
ASSET_CLOSE_TO_KEY = "AssetCloseTo"
SENDER_KEY = "Sender"

TX_FIELDS = [
    REKEY_TO_KEY,
    CLOSE_REMAINDER_TO_KEY,
    ASSET_CLOSE_TO_KEY,
    SENDER_KEY,
]

# used to represent universal address set
ANY_ADDRESS = "ANY_ADDRESS"
# used to represent null address set
NO_ADDRESS = "NO_ADDRESS"
# used to represent single unknown address.
SOME_ADDRESS = "SOME_ADDRESS"
# used to represent application's creator address returned by `global CreatorAddress`
# distinguishing creator address will be helpful for the fuzzer (?)
CREATOR_ADDRESS = "CREATOR_ADDRESS"


class AddrFields(DataflowTransactionContext):  # pylint: disable=too-few-public-methods

    BASE_KEYS: List[str] = TX_FIELDS
    KEYS_WITH_GTXN: List[str] = TX_FIELDS

    def _universal_set(self, key: str = "") -> Set:
        return set([ANY_ADDRESS])

    def _null_set(self, key: str = "") -> Set:
        return set([NO_ADDRESS])

    def _union(self, key: str, a: Set, b: Set) -> Set:
        """A union U = U, A union NullSet = A

        Args:
            key: The analysis key. The values in set :a: and :b: are values for this key.
            a: Set 1.
            b: Set 2.

        Returns:
            Returns union of set a and b.
        """
        if ANY_ADDRESS in a or ANY_ADDRESS in b:
            return self._universal_set()

        if NO_ADDRESS in a and NO_ADDRESS in b:
            return self._null_set()
        if NO_ADDRESS in a:
            return b
        if NO_ADDRESS in b:
            return a

        return a | b

    def _intersection(self, key: str, a: Set, b: Set) -> Set:
        """A intersection NullSet = NullSet, A intersection U = A

        Args:
            key: The analysis key. The values in set :a: and :b: are values for this key.
            a: Set 1.
            b: Set 2.

        Returns:
            Returns union of set a and b.
        """
        if NO_ADDRESS in a or NO_ADDRESS in b:
            return self._null_set()

        if ANY_ADDRESS in a and ANY_ADDRESS in b:
            return self._universal_set()
        if ANY_ADDRESS in a:
            return set(b)
        if ANY_ADDRESS in b:
            return set(a)

        return a & b

    def _get_asserted_address(self, ins: "Instruction") -> Set[str]:
        """return set of address which are represented by the ins

        ZeroAddress is considered to be null set.

        Args:
            ins: An instruction. The function considers that executing :ins: pushes an
                "address" value onto the stack. Based on the instruction, the function
                returns list of possible values.

        Returns:
            Returns a list of possible address values for the field when that field value is
            asserted against the output of :ins:.
        """
        if isinstance(ins, Global) and isinstance(ins.field, ZeroAddress):
            # ZeroAddress
            return self._null_set()
        if isinstance(ins, Addr):
            if ins.addr == ZERO_ADDRESS:
                return self._null_set()
            return set([ins.addr])
        if isinstance(ins, Global) and isinstance(ins.field, CreatorAddress):
            return set([CREATOR_ADDRESS])
        # return SOME_ADDRESS here
        # we don't know the address this particular instruction will return.
        # But it will only return a single address even though we don't know what it is.
        return set([f"{SOME_ADDRESS}_{ins}"])

    def _get_asserted_txn_gtxn(
        self, key: str, ins_stack_value: KnownStackValue
    ) -> Tuple[Set[str], Set[str]]:
        if not isinstance(ins_stack_value.instruction, (Eq, Neq)):
            return self._universal_set(), self._universal_set()

        arg1 = ins_stack_value.args[0]
        arg2 = ins_stack_value.args[1]

        asserted_addresses = None
        if isinstance(arg1, UnknownStackValue) and isinstance(arg2, UnknownStackValue):
            # arg1 and arg2 are unknown value.
            return self._universal_set(), self._universal_set()

        if isinstance(arg1, UnknownStackValue):
            if not isinstance(arg2, UnknownStackValue) and not is_value_matches_key(key, arg2):
                # arg1 is unknown and arg2 is not related to "key"
                return self._universal_set(), self._universal_set()
            # arg2 is related to "key" but arg1 is unknown
            asserted_addresses = set([SOME_ADDRESS])
        elif isinstance(arg2, UnknownStackValue):
            if not isinstance(arg1, UnknownStackValue) and not is_value_matches_key(key, arg1):
                # arg2 is unknown and arg1 is not related to "key"
                return self._universal_set(), self._universal_set()
            # arg1 is related to "key" but arg2 is unknown
            asserted_addresses = set([SOME_ADDRESS])
        elif is_value_matches_key(key, arg1):
            asserted_addresses = self._get_asserted_address(arg2.instruction)
        elif is_value_matches_key(key, arg2):
            asserted_addresses = self._get_asserted_address(arg1.instruction)

        if asserted_addresses is None:
            return self._universal_set(), self._universal_set()

        if isinstance(ins_stack_value.instruction, Eq):
            return asserted_addresses, self._universal_set()
        return self._universal_set(), asserted_addresses

    def _get_asserted_single(self, key: str, ins_stack_value: KnownStackValue) -> Tuple[Set, Set]:
        return self._get_asserted_txn_gtxn(key, ins_stack_value)

    @staticmethod
    def _set_addr_values(ctx_addr_value: "AddrFieldValue", addr_values: Set[str]) -> None:
        ctx_addr_value.any_addr = ANY_ADDRESS in addr_values
        ctx_addr_value.no_addr = NO_ADDRESS in addr_values
        ctx_addr_value.possible_addr = list(addr_values - set([ANY_ADDRESS, NO_ADDRESS]))

    def _store_results(self) -> None:
        key_and_addr_obj: List[
            Tuple[str, Callable[["BlockTransactionContext"], "AddrFieldValue"]]
        ] = [
            (REKEY_TO_KEY, lambda ctx: ctx.rekeyto),
            (CLOSE_REMAINDER_TO_KEY, lambda ctx: ctx.closeto),
            (ASSET_CLOSE_TO_KEY, lambda ctx: ctx.assetcloseto),
            (SENDER_KEY, lambda ctx: ctx.sender),
        ]
        for key, addr_field_obj in key_and_addr_obj:
            if key not in self.BASE_KEYS:
                continue
            for block in self._function.blocks:
                self._set_addr_values(
                    addr_field_obj(self._function.transaction_context(block)),
                    self._block_contexts[key][block],
                )
                for idx in range(16):
                    addr_values = self._block_contexts[get_gtxn_at_index_key(idx, key)][block]
                    self._set_addr_values(
                        addr_field_obj(self._function.transaction_context(block).gtxn_context(idx)),
                        addr_values,
                    )

                    abs_addr_values = self._block_contexts[get_absolute_index_key(idx, key)][block]
                    self._set_addr_values(
                        addr_field_obj(
                            self._function.transaction_context(block).absolute_context(idx)
                        ),
                        abs_addr_values,
                    )

                for offset in range(-(MAX_GROUP_SIZE - 1), MAX_GROUP_SIZE):
                    if offset == 0:
                        continue
                    rel_addr_values = self._block_contexts[get_relative_index_key(offset, key)][
                        block
                    ]
                    self._set_addr_values(
                        addr_field_obj(
                            self._function.transaction_context(block).relative_context(offset)
                        ),
                        rel_addr_values,
                    )
